/******************************************************************

	MUSCLE SmartCard Development ( http://www.linuxnet.com )
	    Title  : eventhandler.c
	    Package: pcsc lite
            Author : David Corcoran
            Date   : 3/13/00
	    License: Copyright (C) 2000 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This keeps track of card insertion/removal events
	    and updates ATR, protocol, and status information.

********************************************************************/

#include <thread_generic.h>
#include <readerfactory.h>
#include <dyn_generic.h>
#include <sys_generic.h>
#include <ifdhandler.h> 
#include <ifdwrapper.h>
#include <debuglog.h>

void EHStatusHandlerThread( PREADER_CONTEXT );

LONG EHDestroyEventHandler( PREADER_CONTEXT rContext ) {

  LONG rv;

  /* Set the thread to 0 to exit thread */
  rContext->dwLockId = 0xFFFF;

  DebugLogA("EHDestroyEventHandler: Stomping thread. \n", __FILE__, __LINE__);

  do {
  /* Wait for the child to respond */
  SYS_USleep(50000);
  } while ( rContext->dwLockId == 0xFFFF );

  DebugLogA("EHDestroyEventHandler: Thread stomped. \n", __FILE__, __LINE__);

  return SCARD_S_SUCCESS;
}

LONG EHSpawnEventHandler( PREADER_CONTEXT rContext ) {

  LONG rv;
  LPCSTR lpcReader;
  DWORD dwStatus, dwProtocol;
  
  /* Zero out everything */
  rv=0; lpcReader=0; dwStatus=0; dwProtocol=0;

  lpcReader = rContext->lpcReader;
  
  rv = IFDStatusICC( rContext, &dwStatus,
		     &dwProtocol, rContext->ucAtr, 
		     &rContext->dwAtrLen );    

  if ( rv != SCARD_S_SUCCESS ) {
    DebugLogC("EHSpawnEventHandler: Initial Check Failed on %s\n", lpcReader, 
	      __FILE__, __LINE__);
    SYS_Exit(1);
  }

  rv = SYS_ThreadCreate( &rContext->pthThread, NULL, 
			 (LPVOID)EHStatusHandlerThread,
			 (LPVOID)rContext );
  if ( rv == 1 ) {
    return SCARD_S_SUCCESS;
  } else {
    return SCARD_E_NO_MEMORY;
  }

}


void EHStatusHandlerThread( PREADER_CONTEXT rContext ) {

  LONG rv;
  LPCSTR lpcReader;
  DWORD dwStatus, dwProtocol;
  DWORD dwErrorCount, dwCurrentState;

  /* Zero out everything */
  rv=0; lpcReader=0; dwStatus=0; dwProtocol=0; 
  dwCurrentState=0; dwErrorCount=0;

  lpcReader = rContext->lpcReader;

  rv = IFDStatusICC( rContext, &dwStatus,
		     &dwProtocol, rContext->ucAtr, 
		     &rContext->dwAtrLen );    
  
  if ( dwStatus & SCARD_PRESENT ) {
    rv = IFDPowerICC( rContext, IFD_POWER_UP, 
		      rContext->ucAtr, 
		      &rContext->dwAtrLen );
    rContext->dwProtocol  = PHGetDefaultProtocol( rContext->ucAtr, 
						  rContext->dwAtrLen );
    rContext->dwStatus   |= SCARD_PRESENT;
    rContext->dwStatus   &= ~SCARD_ABSENT;
    rContext->dwStatus   |= SCARD_POWERED;
    rContext->dwStatus   |= SCARD_NEGOTIABLE;   
    dwCurrentState        = SCARD_PRESENT;
  } else {
    dwCurrentState        = SCARD_ABSENT;
    rContext->dwStatus   |= SCARD_ABSENT;
    rContext->dwStatus   &= ~SCARD_PRESENT;
    rContext->dwAtrLen    = 0;
    rContext->dwProtocol  = 0;
  }

  while(1) {      

    dwStatus   = 0;
    
    rv = IFDStatusICC( rContext, &dwStatus,
		       &dwProtocol, rContext->ucAtr, 
		       &rContext->dwAtrLen );
    
    if ( rv != SCARD_S_SUCCESS ) {
      DebugLogC("EHSpawnEventHandler: Error communicating to: %s\n",
		lpcReader, __FILE__, __LINE__ );

      /* This code causes race conditions on 
	 G4's with USB insertion

      dwErrorCount += 1;
      if ( dwErrorCount == 5 ) {
	RFUnInitializeReader( rContext );
	RFInitializeReader( rContext );
	dwErrorCount = 0;
	SYS_Sleep(3);
	}

      End of race condition code block */

    }


    if ( dwStatus & SCARD_ABSENT ) {
      if ( dwCurrentState == SCARD_PRESENT ) {
	
	/* Change the status structure */
#ifdef PCSC_DEBUG
	DebugLogC("EHSpawnEventHandler: Card Removed From %s\n", 
		  lpcReader, __FILE__, __LINE__);
#endif	  
	/* Notify the card has been removed */
	RFSetReaderEventState( rContext, SCARD_REMOVED );

	rContext->dwAtrLen   = 0;
	rContext->dwProtocol = 0;
	rContext->dwStatus  |= SCARD_ABSENT;
	rContext->dwStatus  &= ~SCARD_PRESENT;
	rContext->dwStatus  &= ~SCARD_POWERED;
	rContext->dwStatus  &= ~SCARD_NEGOTIABLE;
	dwCurrentState       = SCARD_ABSENT;
      }
      
    } else if ( dwStatus & SCARD_PRESENT ) {
      if ( dwCurrentState == SCARD_ABSENT ) {
	/* Power and reset the card */
	SYS_USleep( PCSCLITE_STATUS_WAIT );
	rv = IFDPowerICC( rContext, IFD_POWER_UP, 
			  rContext->ucAtr, 
			  &rContext->dwAtrLen );

	/* Notify the card has been reset */
	RFSetReaderEventState( rContext, SCARD_RESET );

	rContext->dwProtocol  = PHGetDefaultProtocol( rContext->ucAtr, 
						      rContext->dwAtrLen );
	rContext->dwStatus   |= SCARD_PRESENT;
	rContext->dwStatus   &= ~SCARD_ABSENT;
	rContext->dwStatus   |= SCARD_POWERED;
	rContext->dwStatus   |= SCARD_NEGOTIABLE; 
	dwCurrentState        = SCARD_PRESENT;

#ifdef PCSC_DEBUG
	DebugLogC("EHSpawnEventHandler: Card inserted into %s\n", 
		  lpcReader, __FILE__, __LINE__);

	if ( rContext->dwAtrLen > 0 ) {
	  DebugLogD("EHSpawnEventHandler: Card insertion ATR:\n\t", 
		    rContext->ucAtr, rContext->dwAtrLen,  __FILE__, __LINE__ );
	} else {
	  DebugLogA("EHSpawnEventHandler: Error powering up card.\n", 
		    __FILE__, __LINE__);
	}	  
#endif
      }
    }     
    
    if ( rContext->dwLockId == 0xFFFF ) {
      /* Exit and notify the caller */
      rContext->dwLockId = 0; 
      SYS_ThreadDetach( rContext->pthThread );
      SYS_ThreadExit(0);
    }

    SYS_USleep( 800000 );
  }
}
 
