/******************************************************************

	MUSCLE SmartCard Development ( http://www.linuxnet.com )
	    Title  : hotplug_macosx.c
	    Package: pcsc lite
            Author : David Corcoran
            Date   : 10/25/00
	    License: Copyright (C) 2000 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This provides a search API for hot pluggble
	             devices.
	            
********************************************************************/

#include <wintypes.h>
#include <pcsclite.h>  
#include <readerfactory.h>
#include <stdio.h>
#include <IOKitLib.h>
#include <USB.h>
#include <USBSpec.h>
#include <IOUSBLib.h>
#include <assert.h>
#include <string.h>
#include <CoreFoundation/CFBundle.h>
#include <CoreFoundation/CFString.h>
#include <CoreFoundation/CFURL.h>

static IOUSBInterfaceRef newInterface;
static IOUSBIteratorRef usbiterator;
static LONG initBundleTracker = 0;

#define PCSCLITE_HP_DROPDIR  "/usr/local/pcsc/drivers/"


/* A list to keep track of 20 simultaneous readers */

static struct _bundleTracker {
  int iStatus;
  int isConnected;
} bundleTracker[20];


LONG HPSearchHotPluggables() {

  LONG rv;
  IOReturn iorv, dorv;
  kern_return_t krv;

  IOUSBFindEndpointRequest epreq;
  IONotificationPortRef ionotify;
  IOUSBConfigurationDescriptorPtr configdesc;
  IOUSBInterfaceDescriptor ifacedesc;
  IOUSBFindInterfaceRequest ifacereq;
  IOUSBDeviceDescriptor desc;
  IOUSBMatch usbMatch;
  IOUSBDeviceRef newDevice;

  CFArrayRef bundleArray;
  CFURLRef pluginDirURL, bundPathURL;
  CFDictionaryRef bundleInfoDict;
  CFStringRef propertyString, pluginDirString, cfStringURL;
  CFBundleRef currBundle;
  LONG hpManu_id, hpProd_id;

  const char *cStringValue, *cStringLibPath;
  int bundleArraySize;
  char *path;

  mach_port_t masterPort, notifyPort;
  UInt8 configsize; 
  int i;


   if ( initBundleTracker == 0 ) {
     for (i=0; i < 20; i++) {
       bundleTracker[i].isConnected = 0;
     }
     initBundleTracker = 1;
   } 
    
    iorv = IOMasterPort( bootstrap_port, &masterPort );
    if ( iorv != 0 ) return -1;
    
    ionotify   = IONotificationPortCreate( masterPort );
    notifyPort = IONotificationPortGetMachPort( ionotify );
    
    path = strdup(PCSCLITE_HP_DROPDIR);
    
    pluginDirString = CFStringCreateWithCString(NULL, path,
						kCFStringEncodingMacRoman);
    
    if (pluginDirString == NULL)
      return -1;
    
    pluginDirURL = CFURLCreateWithFileSystemPath(NULL, pluginDirString,
						 kCFURLPOSIXPathStyle, TRUE);
    if (pluginDirURL == NULL)
      return -1;

  /* CFRelease(pluginDirURL); */

  bundleArray = CFBundleCreateBundlesFromDirectory( NULL, pluginDirURL, NULL );

  if ( bundleArray == NULL ) 
    return SCARD_E_NO_MEMORY;
  
  bundleArraySize = CFArrayGetCount( bundleArray );
  
  for (i=0; i<bundleArraySize; i++) {
    currBundle = (CFBundleRef)CFArrayGetValueAtIndex( bundleArray, i );
    bundleInfoDict = CFBundleGetInfoDictionary( currBundle );
    if ( bundleInfoDict == NULL )
      return SCARD_E_NO_MEMORY;   
    
    propertyString = CFDictionaryGetValue( bundleInfoDict, 
					   CFSTR("ifdVendorID") );
    if ( propertyString == 0 )
      return 0;
    
    cStringValue = CFStringGetCStringPtr( propertyString, 
					  CFStringGetSystemEncoding() );
    hpManu_id = strtoul( cStringValue, 0, 16 );
    
    propertyString = CFDictionaryGetValue( bundleInfoDict, 
					   CFSTR("ifdProductID") );
    if ( propertyString == 0 )
      return 0;
    
    cStringValue = CFStringGetCStringPtr( propertyString, 
					  CFStringGetSystemEncoding() );
    hpProd_id = strtoul( cStringValue, 0, 16 );
    
    usbMatch.usbClass    = kIOUSBAnyClass;
    usbMatch.usbSubClass = kIOUSBAnySubClass;
    usbMatch.usbProtocol = kIOUSBAnyProtocol;
    usbMatch.usbVendor   = hpManu_id;
    usbMatch.usbProduct  = hpProd_id;
    
    /* Only concerned about devices */
    krv = IOUSBCreateIterator( masterPort, notifyPort, &usbMatch, 
			       &usbiterator );
    krv = IOUSBIteratorNext( usbiterator );
    IOUSBDisposeIterator( usbiterator );
    
    if ( krv == 0 ) {
      if ( bundleTracker[i].isConnected == 0 ) {
	fprintf(stderr, "Adding Reader.\n");
	
	propertyString = CFDictionaryGetValue( bundleInfoDict, 
					       CFSTR("ifdFriendlyName"));
	if ( propertyString == 0 )
	  return 0;
	
	bundPathURL    = CFBundleCopyBundleURL( currBundle );
	cfStringURL    = CFURLCopyPath(  bundPathURL );
	cStringLibPath = strdup(CFStringGetCStringPtr( cfStringURL, 
						CFStringGetSystemEncoding()));
	cStringValue   = strdup(CFStringGetCStringPtr( propertyString, 
						CFStringGetSystemEncoding()));
	
	rv = RFAddReader( (LPSTR)cStringValue, 0x000001, 
                          (LPSTR)cStringLibPath );
	
	free(cStringValue);
	free(cStringLibPath);
	
	if ( rv != SCARD_S_SUCCESS ) {
	  return 0;
	}
	
	bundleTracker[i].isConnected = 1;
      }
    } else {
      if ( bundleTracker[i].isConnected == 1 ) {
	fprintf(stderr, "Disconnecting Reader\n");
	
	propertyString = CFDictionaryGetValue( bundleInfoDict, 
					       CFSTR("ifdFriendlyName"));
	if ( propertyString == 0 )
	  return 0;
	
	cStringValue = strdup(CFStringGetCStringPtr( propertyString, 
						CFStringGetSystemEncoding()));
	
	rv = RFRemoveReader( (LPSTR)cStringValue );
	
	if ( rv != SCARD_S_SUCCESS ) {
	  return 0;
	}
	
	bundleTracker[i].isConnected = 0;
      }
    }
    
  } /* End of for ... */

  return 1;
}

