/******************************************************************

	MUSCLE SmartCard Development ( http://www.linuxnet.com )
	    Title  : readerfactory.c
	    Package: pcsc lite
            Author : David Corcoran
            Date   : 7/27/99
	    License: Copyright (C) 1999 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This keeps track of a list of currently 
	             available reader structures.

********************************************************************/

#include <readerfactory.h>
#include <dyn_generic.h>
#include <sys_generic.h>
#include <eventhandler.h>
#include <ifdhandler.h>
#include <ifdwrapper.h>
/* FIX :: This should go */
#include <fcntl.h>
#include <debuglog.h>
#include <stdlib.h>
#include <string.h>

static PREADER_CONTEXT sContexts[PCSCLITE_MAX_CONTEXTS];
static DWORD *dwNumContexts = 0;
static struct _sHandleStruct {
  UCHAR ucInUse;
} sHandleStruct[PCSCLITE_MAX_CONTEXTS*PCSCLITE_MAX_CONTEXTS];

LONG RFAllocateReaderSpace( DWORD dwAllocNum ) {

  int psize;                  /* Page Size       */
  int fd;                     /* File descriptor */
  int i;                      /* Counter         */
  int rv;                     /* Return tester   */

  /* Zero out everything */
  i=0; fd=0; psize=0;

  /* Here we must map the sContexts over to a memory map so when
     the server is threaded or forked we can continue to keep 
     track of the reader's state.
  */
  
  /* First open and lock the file to test for other running daemons */

  fd = SYS_OpenFile( PCSCLITE_SHM_FILE, O_RDWR | O_CREAT, 0644 );
  if ( fd < 0 ) {
    DebugLogA("Error: Cannot open shared file\n", __FILE__, __LINE__);
    SYS_Exit(1);
  }

  /* Attempt a passive lock on the shared memory file */
  rv = SYS_LockFile( fd );

  if ( rv != 0 ) {
      DebugLogA("Error: pcscd already running\n", __FILE__, __LINE__);
      SYS_Exit(1);
  }

  /* Lock will be removed on close */
  SYS_CloseFile( fd );

  /* Attempt to remove an already existing shared memory file */
  SYS_RemoveFile( PCSCLITE_SHM_FILE );
  fd = SYS_OpenFile( PCSCLITE_SHM_FILE, O_RDWR | O_CREAT, 0600 );
  
  if ( fd < 0 ) {
    DebugLogA("Error: Cannot open shared file\n", __FILE__, __LINE__);
    SYS_Exit(1);
  }
  
  /* Lock the file again - should work since above did not fail */
  SYS_LockFile( fd );
  
  /* Shared segments must be on page borders */
  psize = SYS_GetPageSize();
  
  /* Jump to end of file space and allocate zero's */
  SYS_SeekFile( fd, psize * (dwAllocNum + 1) );    
  SYS_WriteFile( fd, "", 1 );
  
  /* Allocate global dwNumContexts */
  dwNumContexts  = (DWORD *)SYS_MemoryMap( sizeof(DWORD), fd, 0 );
  *dwNumContexts = 0;
  
  /* Allocate each reader structure */
  for ( i=0; i < dwAllocNum; i++ ) {
    sContexts[i] = (PREADER_CONTEXT)
      SYS_MemoryMap( sizeof(READER_CONTEXT)*dwAllocNum, 
		     fd, (i+1)*(psize) );
    
  }
  
  /* Now initialize the handle structure */
  for ( i=0; i < PCSCLITE_MAX_CONTEXTS*PCSCLITE_MAX_CONTEXTS; i++ ) {
    sHandleStruct[i].ucInUse = 0;
  }
}

LONG RFAddReader( LPSTR lpcReader, DWORD dwPort, LPSTR lpcLibrary ) {
  
  DWORD dwContext;
  LONG rv;

  static int fd = 0;
  int i, psize;

  /* Zero out everything */
  dwContext=0; rv=0; i=0; psize=0;

  if ( lpcReader == 0 || lpcLibrary == 0 ) {
    return SCARD_E_INVALID_VALUE;
  }

  if ( dwNumContexts != 0 ) {
    for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
      if ( (sContexts[i])->vHandle != 0 ) {
        if ((strcmp(lpcReader, (sContexts[i])->lpcReader) == 0)) {
	  return SCARD_E_DUPLICATE_READER;
        }
      }
    }
  }

  /* We must find an empty spot to put the reader structure */
  for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
    if ( (sContexts[i])->vHandle == 0 ) {
      dwContext = i;
      break;
    }
  }
  
  if ( i == PCSCLITE_MAX_CONTEXTS ) {
    /* No more spots left return */
    return SCARD_E_NO_MEMORY;
  }

  strcpy((sContexts[dwContext])->lpcReader, lpcReader);
  strcpy((sContexts[dwContext])->lpcLibrary, lpcLibrary);
  (sContexts[dwContext])->dwVersion       = 0;
  (sContexts[dwContext])->dwPort          = dwPort;
  (sContexts[dwContext])->mMutex          = 0;
  (sContexts[dwContext])->dwStatus        = 0;
  (sContexts[dwContext])->dwBlockStatus   = 0;
  (sContexts[dwContext])->dwContexts      = 0;
  (sContexts[dwContext])->dwSlot          = 1;
  (sContexts[dwContext])->dwLockId        = 0;
  (sContexts[dwContext])->vHandle         = 0;
  (sContexts[dwContext])->dwIdentity      = 
    (dwContext + 1) << (sizeof(DWORD) / 2) * 8;

  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    (sContexts[dwContext])->psHandles[i].hCard = 0;
  }

  /* Check to see if an existing reader is specified on the
     same port and uses the same library - if that is true
     then it must be a reader that supports multiple slots  */

  if ( *dwNumContexts != 0 ) {
    for ( i=0; i < dwContext; i++ ) {
      if (((strcmp(sContexts[i]->lpcLibrary, lpcLibrary) == 0)&&
	   (sContexts[i])->dwPort == dwPort)) {
	if ( (sContexts[i])->dwSlot == (sContexts[dwContext])->dwSlot ) {
	  (sContexts[dwContext])->dwSlot += 1;
	  (sContexts[dwContext])->mMutex  = (sContexts[i])->mMutex;
	}     
      }
    }
  }

  if ( (sContexts[dwContext])->mMutex == 0 ) {
    (sContexts[dwContext])->mMutex = 
      (PCSCLITE_MUTEX_T)malloc(sizeof(PCSCLITE_MUTEX));
    SYS_MutexInit( (sContexts[dwContext])->mMutex );
  }

  *dwNumContexts += 1;

  rv = RFInitializeReader( sContexts[dwContext] );
  if ( rv != SCARD_S_SUCCESS ) {
    /* Cannot connect to reader exit gracefully */
    DebugLogC("RFAddReader: %s init failed.\n", lpcReader, 
	      __FILE__, __LINE__ );

    /* This code piece caused seg fault 
       rv = RFUnInitializeReader( sContexts[dwContext] );
       End of segmentation fault */
  }

  EHSpawnEventHandler( sContexts[dwContext] );

  return SCARD_S_SUCCESS;
}

LONG RFRemoveReader( LPSTR lpcReader ) {
  
  LONG rv;
  PREADER_CONTEXT sContext;
  int i;

  /* Zero out everything */
  i=0;

  if ( lpcReader == 0 ) {
    return SCARD_E_INVALID_VALUE;
  }

  rv = RFReaderInfo( lpcReader, &sContext );
  if ( rv != SCARD_S_SUCCESS ) {
    return rv;
  }

  /* Try to destroy the thread */
  rv = EHDestroyEventHandler( sContext );

  rv = RFUnInitializeReader( sContext );
  if ( rv != SCARD_S_SUCCESS ) {
    return rv;
  }

  /* Destroy and free the mutex */
  SYS_MutexDestroy( sContext->mMutex );
  free( sContext->mMutex );

  sContext->dwVersion       = 0;
  sContext->dwPort          = 0;
  sContext->mMutex          = 0;
  sContext->dwStatus        = 0;
  sContext->dwBlockStatus   = 0;
  sContext->dwContexts      = 0;
  sContext->dwSlot          = 0;
  sContext->dwLockId        = 0;
  sContext->vHandle         = 0;
  sContext->dwIdentity      = 0;

  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    sContext->psHandles[i].hCard = 0;
  }
  
  /* We might have to swap device allocators        */
  /* Take the last one and swap it with the current */
  
  *dwNumContexts -= 1;

  return SCARD_S_SUCCESS;
}

LONG RFListReaders( LPSTR lpcReaders, LPDWORD pdwReaderNum ) {

  DWORD dwCSize;
  LPSTR lpcTReaders;
  int i, p;

  /* Zero out everything */
  dwCSize = 0; lpcTReaders = 0; i=0; p=0;

  if ( *dwNumContexts == 0 ) {
    return SCARD_E_READER_UNAVAILABLE;
  }

  /* Ignore the groups for now, return all readers */
  for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
    if ( (sContexts[i])->vHandle != 0 ) {
      dwCSize += strlen( (sContexts[i])->lpcReader ) + 1;
      p += 1;
    }
  }
  
  if ( p > *dwNumContexts ) {
    /* We are severely hosed here        */
    /* Hopefully this will never be true */
    SCARD_F_UNKNOWN_ERROR;
   }
  
  /* Added for extra NULL byte on MultiString */
  dwCSize += 1;  

  /* If lpcReaders is not allocated then just */
  /* return the amount needed to allocate     */
  if ( lpcReaders == 0 ) {
    *pdwReaderNum = dwCSize;
    return SCARD_S_SUCCESS;
  }
 
  if ( *pdwReaderNum < dwCSize ) {
    return SCARD_E_INSUFFICIENT_BUFFER;
  }

  *pdwReaderNum = dwCSize;
  lpcTReaders   = lpcReaders;
  p = 0;
  
  /* Creating MultiString */
  for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
    if ( (sContexts[i])->vHandle != 0 ) {
      strcpy(&lpcTReaders[p], (sContexts[i])->lpcReader );
      p += strlen((sContexts[i])->lpcReader);              /* Copy       */
      lpcTReaders[p] = 0;                                  /* Add NULL   */
      p += 1;                                              /* Move on    */
    }
  }
  
  lpcTReaders[p] = 0;                                    /* Add NULL   */

  return SCARD_S_SUCCESS;
}  

LONG RFReaderInfo( LPSTR lpcReader, PREADER_CONTEXT* sReader ) {
  
  int i; 

  /* Zero out everything */
  i=0;

  if ( lpcReader == 0 ) {
    return SCARD_E_INVALID_VALUE;
  }

  for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
    if ( (sContexts[i])->vHandle != 0 ) {
      if (strcmp(lpcReader, (sContexts[i])->lpcReader) == 0) {
        *sReader = sContexts[i];
        return SCARD_S_SUCCESS;
      }
    }
  }  

  return SCARD_E_INVALID_VALUE;
}

LONG RFReaderInfoById( DWORD dwIdentity, PREADER_CONTEXT* sReader ) {
  
  int i;

  /* Zero out everything */
  i=0;

  /* Strip off the lower nibble and get the identity */

  dwIdentity = dwIdentity >> (sizeof(DWORD)/2)*8;
  dwIdentity = dwIdentity << (sizeof(DWORD)/2)*8;

  for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
    if ( dwIdentity == (sContexts[i])->dwIdentity ) {
      *sReader = sContexts[i];
      return SCARD_S_SUCCESS;
    }
  }

  return SCARD_E_INVALID_VALUE;
}

LONG RFLoadReader( PREADER_CONTEXT rContext ) {

  LONG rv;

  if ( rContext->vHandle != 0 ) {
    DebugLogA("RFLoadReader: Warning library pointer not NULL\n", __FILE__,
	      __LINE__ );
  }

  rv = DYN_LoadLibrary( &rContext->vHandle, rContext->lpcLibrary );
  return rv;
}

LONG RFBindFunctions( PREADER_CONTEXT rContext ) {

  LONG rv, ret;
  LPVOID pvTestA, pvTestB;

  /* Zero out everything */
  rv=0; ret=0; pvTestA=0; pvTestB;


  /* Use this function as a dummy to determine the IFD
     Handler version type.  1.0/2.0.  Suppress error
     messaging since it can't be 1.0 and 2.0.
  */

  DebugLogSuppress( DEBUGLOG_IGNORE_ENTRIES );

  rv = DYN_GetAddress( rContext->vHandle, 
		       &rContext->psFunctions.pvfCreateChannel, 
		       "IO_Create_Channel" );
  
  ret = DYN_GetAddress( rContext->vHandle, 
			&rContext->psFunctions.pvfCreateChannel, 
			"IFDHCreateChannel" );

  DebugLogSuppress( DEBUGLOG_LOG_ENTRIES );

  if ( rv != SCARD_S_SUCCESS && ret != SCARD_S_SUCCESS ) {
    /* Neither version of the IFD Handler was found - exit */
    rContext->psFunctions.pvfCreateChannel = 0;
    SYS_Exit(1);
  } else if ( rv == SCARD_S_SUCCESS ) {
    /* Ifd Handler 1.0 found                                */
    /* Re bind the function since it was lost in the second */
    rContext->dwVersion |= IFD_HVERSION_1_0;
    DYN_GetAddress( rContext->vHandle, 
		    &rContext->psFunctions.pvfCreateChannel, 
		    "IO_Create_Channel" );
  } else {
    /* Ifd Handler 2.0 found */
    rContext->dwVersion |= IFD_HVERSION_2_0;
  }



  /* The following binds version 1.0 of the IFD Handler specs */

  if ( rContext->dwVersion & IFD_HVERSION_1_0 ) {
    
    DebugLogA("RFBindFunctions: Loading IFD Handler 1.0\n", __FILE__,
	      __LINE__ );

    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfCloseChannel, 
			 "IO_Close_Channel" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfCloseChannel = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfGetCapabilities, 
			 "IFD_Get_Capabilities" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfGetCapabilities = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfSetCapabilities, 
			 "IFD_Set_Capabilities" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfSetCapabilities = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfSetProtocol, 
			 "IFD_Set_Protocol_Parameters" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfSetProtocol = 0;
      /* Not a completely required function */
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfPowerICC, 
			 "IFD_Power_ICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfPowerICC = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfSwallowICC, 
			 "IFD_Swallow_ICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfSwallowICC = 0;
      /* Not a completely required function */
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfEjectICC, 
			 "IFD_Eject_ICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfEjectICC = 0;
      /* Not a completely required function */
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfConfiscateICC, 
			 "IFD_Confiscate_ICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfConfiscateICC = 0;
      /* Not a completely required function */
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfTransmitICC, 
			 "IFD_Transmit_to_ICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfTransmitICC = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfICCPresent, 
			 "IFD_Is_ICC_Present" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfICCPresent = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfICCAbsent, 
			 "IFD_Is_ICC_Absent" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfICCAbsent = 0;
      /* Not a completely required function */
    }


  /* The following binds version 2.0 of the IFD Handler specs */

  } else if ( rContext->dwVersion & IFD_HVERSION_2_0 ) {

    DebugLogA("RFBindFunctions: Loading IFD Handler 2.0\n", __FILE__,
	      __LINE__ );

    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfCloseChannel, 
			 "IFDHCloseChannel" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfCloseChannel = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfGetCapabilities, 
			 "IFDHGetCapabilities" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfGetCapabilities = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfSetCapabilities, 
			 "IFDHSetCapabilities" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfSetCapabilities = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfSetProtocol, 
			 "IFDHSetProtocolParameters" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfSetProtocol = 0;
      /* Not a completely required function */
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfPowerICC, 
			 "IFDHPowerICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfPowerICC = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfTransmitICC, 
			 "IFDHTransmitToICC" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfTransmitICC = 0;
      SYS_Exit(1);
    }
    
    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfControl, 
			 "IFDHControl" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfControl = 0;
      SYS_Exit(1);
    }

    rv = DYN_GetAddress( rContext->vHandle, 
			 &rContext->psFunctions.pvfICCPresent, 
			 "IFDHICCPresence" );
    
    if ( rv != SCARD_S_SUCCESS ) {
      rContext->psFunctions.pvfICCPresent = 0;
      SYS_Exit(1);
    }
    
  } else {
    /* Who knows what could have happenned for it to get here. */
    DebugLogA("RFBindFunctions: IFD Handler not 1.0/2.0\n", __FILE__,
	      __LINE__ );
    SYS_Exit(1);
  }


  return SCARD_S_SUCCESS;
}

LONG RFUnBindFunctions( PREADER_CONTEXT rContext ) {

  /* Zero out everything */
  
  rContext->psFunctions.pvfCreateChannel   = 0;
  rContext->psFunctions.pvfCloseChannel    = 0;
  rContext->psFunctions.pvfGetCapabilities = 0;
  rContext->psFunctions.pvfSetCapabilities = 0;
  rContext->psFunctions.pvfSetProtocol     = 0;
  rContext->psFunctions.pvfPowerICC        = 0;
  rContext->psFunctions.pvfSwallowICC      = 0;
  rContext->psFunctions.pvfEjectICC        = 0;
  rContext->psFunctions.pvfConfiscateICC   = 0;
  rContext->psFunctions.pvfTransmitICC     = 0;
  rContext->psFunctions.pvfICCPresent      = 0;
  rContext->psFunctions.pvfICCAbsent       = 0;

  return SCARD_S_SUCCESS;
}

LONG RFUnloadReader( PREADER_CONTEXT rContext ) {
  
  DYN_CloseLibrary( &rContext->vHandle );
  rContext->vHandle = 0;

  return SCARD_S_SUCCESS;
}

LONG RFCheckSharing( DWORD hCard ) {

  DWORD dwSharing;
  PREADER_CONTEXT rContext;

  /* Zero out everything */
  dwSharing=0; rContext=0;

  RFReaderInfoById(hCard, &rContext);
  dwSharing = hCard;

  if ( rContext->dwLockId == 0 || 
       rContext->dwLockId == dwSharing ) {
    return SCARD_S_SUCCESS;
  } else {
    return SCARD_E_SHARING_VIOLATION;
  }

}

LONG RFLockSharing( DWORD hCard ) {
  
  PREADER_CONTEXT rContext;

  /* Zero out everything */
  rContext=0;

  RFReaderInfoById(hCard, &rContext);

  if ( RFCheckSharing(hCard) == SCARD_S_SUCCESS ) {
    rContext->dwLockId = hCard;
  } else {
    return SCARD_E_SHARING_VIOLATION;
  }

  return SCARD_S_SUCCESS;
}

LONG RFUnlockSharing( DWORD hCard ) {

  PREADER_CONTEXT rContext;

  /* Zero out everything */
  rContext=0;

  RFReaderInfoById(hCard, &rContext);

  if ( RFCheckSharing(hCard) == SCARD_S_SUCCESS ) {
    rContext->dwLockId = 0;
  } else {
    return SCARD_E_SHARING_VIOLATION;
  }

  return SCARD_S_SUCCESS;
}

LONG RFUnblockContext( SCARDCONTEXT hContext ) {

  int i;

  /* Zero out everything */
  i=0;

  for ( i=0; i < PCSCLITE_MAX_CONTEXTS; i++ ) {
    (sContexts[i])->dwBlockStatus = hContext;
  }
  
  return SCARD_S_SUCCESS;
}

LONG RFUnblockReader( PREADER_CONTEXT rContext ) {

  rContext->dwBlockStatus = BLOCK_STATUS_RESUME;
  return SCARD_S_SUCCESS;
}

LONG RFInitializeReader( PREADER_CONTEXT rContext ) {

  LONG rv;

  /* Zero out everything */
  rv=0;


  /* Spawn the event handler thread */
  DebugLogC("RFInitializeReader: Attempting startup of %s.\n", 
	    rContext->lpcReader, __FILE__, __LINE__);

  /******************************************/
  /* This section loads the library         */
  /******************************************/

  rv = RFLoadReader( rContext );

  if ( rv != SCARD_S_SUCCESS ) { 
    return rv; 
  }
  
  /*******************************************/
  /* This section binds the functions        */
  /*******************************************/

  rv = RFBindFunctions( rContext );

  if ( rv != SCARD_S_SUCCESS ) { 
    RFUnloadReader( rContext );
    return rv; 
  }

  /*******************************************/
  /* This section tries to open the port     */
  /*******************************************/
  
  rv = IFDOpenIFD( rContext, rContext->dwPort );

  if ( rv != IFD_SUCCESS ) { 
#ifdef PCSC_DEBUG
    DebugLogB("RFInitializeReader: Open Port %x Failed\n", rContext->dwPort, 
	      __FILE__, __LINE__);
#endif
    RFUnBindFunctions( rContext );
    RFUnloadReader( rContext );
    return SCARD_E_INVALID_TARGET; 
  }    
  
  return SCARD_S_SUCCESS;
}

LONG RFUnInitializeReader( PREADER_CONTEXT rContext ) { 

  UCHAR Atr[MAX_ATR_SIZE];
  DWORD AtrLen;

  DebugLogC("RFUninitializeReader: Attempting shutdown of %s.\n", 
            rContext->lpcReader, __FILE__, __LINE__);

  /* Close the port, unbind the functions, and unload the library */

  IFDPowerICC( rContext, IFD_POWER_DOWN, Atr, &AtrLen );
  IFDCloseIFD( rContext );
  RFUnBindFunctions( rContext );
  RFUnloadReader( rContext );

  return SCARD_S_SUCCESS;
}

SCARDHANDLE RFCreateReaderHandle( PREADER_CONTEXT rContext ) {

  DWORD lMaxConnections;
  int i;

  /* Zero out everything */
  lMaxConnections=0; i=0;

  lMaxConnections = PCSCLITE_MAX_CONTEXTS*PCSCLITE_MAX_CONTEXTS;

  for ( i=0; i < lMaxConnections; i++ ) {
    if ( sHandleStruct[i].ucInUse == 0 ) {
      sHandleStruct[i].ucInUse = 1;
      break;
    }
  }

  if ( i >= lMaxConnections ) {
    DebugLogB("RFCreateReaderHandle: Too many open handles: %d\n", 
	      lMaxConnections, __FILE__, __LINE__);
  }

  return rContext->dwIdentity + i;
}

LONG RFDestroyReaderHandle( SCARDHANDLE hCard ) {

  /* Pull off the right most half */
  hCard = hCard << (sizeof(DWORD)/2)*8;
  hCard = hCard >> (sizeof(DWORD)/2)*8;

  if ( hCard >= PCSCLITE_MAX_CONTEXTS*PCSCLITE_MAX_CONTEXTS ) {
    DebugLogB("RFDestroyReaderHandle: Out of bounds handle %x\n",
	      hCard, __FILE__, __LINE__);
    return SCARD_E_INVALID_HANDLE;
  }

  sHandleStruct[hCard].ucInUse = 0; 
  return SCARD_S_SUCCESS;
}

LONG RFAddReaderHandle( PREADER_CONTEXT rContext, SCARDHANDLE hCard ) {

  int i;
  i=0;

  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    if ( rContext->psHandles[i].hCard == 0 ) {
      rContext->psHandles[i].hCard         = hCard;
      rContext->psHandles[i].dwEventStatus = 0;
      break;
    }
  }

  if ( i == PCSCLITE_MAX_CONTEXTS ) {     /* List is full */
    return SCARD_E_INSUFFICIENT_BUFFER;
  }

  return SCARD_S_SUCCESS;
}

LONG RFRemoveReaderHandle( PREADER_CONTEXT rContext, SCARDHANDLE hCard ) {

  int i;
  i=0;

  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    if ( rContext->psHandles[i].hCard == hCard ) {
      rContext->psHandles[i].hCard         = 0;
      rContext->psHandles[i].dwEventStatus = 0;
      break;
    }
  }

  if ( i == PCSCLITE_MAX_CONTEXTS ) {     /* Not Found */
    return SCARD_E_INVALID_HANDLE;
  }

  return SCARD_S_SUCCESS;
}

LONG RFSetReaderEventState( PREADER_CONTEXT rContext, DWORD dwEvent ) {

  int i;
  i=0;

  /* Set all the handles for that reader to the event */
  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    if ( rContext->psHandles[i].hCard != 0 ) {
      rContext->psHandles[i].dwEventStatus = dwEvent;
    }
  }

  return SCARD_S_SUCCESS;
}

LONG RFCheckReaderEventState( PREADER_CONTEXT rContext, SCARDHANDLE hCard ) {

  int i;
  i=0;

  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    if ( rContext->psHandles[i].hCard == hCard ) {
      if ( rContext->psHandles[i].dwEventStatus == SCARD_REMOVED ) {
	return SCARD_W_REMOVED_CARD;
      } else if ( rContext->psHandles[i].dwEventStatus == SCARD_RESET ) {
	return SCARD_W_RESET_CARD;
      } else if ( rContext->psHandles[i].dwEventStatus == 0 )  {
	return SCARD_S_SUCCESS;
      } else {
	return SCARD_E_INVALID_VALUE;
      }
    }
  }

  if ( i == PCSCLITE_MAX_CONTEXTS ) {     /* Not Found */
    return SCARD_E_INVALID_HANDLE;
  }
}

LONG RFClearReaderEventState( PREADER_CONTEXT rContext, SCARDHANDLE hCard ) {

  int i;
  i=0;

  for (i=0; i < PCSCLITE_MAX_CONTEXTS; i++) {
    if ( rContext->psHandles[i].hCard == hCard ) {
      rContext->psHandles[i].dwEventStatus = 0;
    }
  }

  if ( i == PCSCLITE_MAX_CONTEXTS ) {     /* Not Found */
    return SCARD_E_INVALID_HANDLE;
  }

  return SCARD_S_SUCCESS;
}
