/******************************************************************

            Title  : pcscdaemon.c
            Package: PC/SC Lite
            Author : David Corcoran
            Date   : 10/24/99
	    License: Copyright (C) 1999 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This is the main pcscd daemon.

********************************************************************/

#include "winscard_rpc.h"
#include <pcsclite.h>
#include <sys_generic.h>
#include <hotplug.h>
#include <sys/errno.h>
#include <stdio.h>

/* Some internal functions */
void SVCService_Run();

/* Some external functions */
extern int  SVCReturnChannelNumber( void );
extern int  SVCReturnProcessNumber( void );
extern void SVCCollectZombies( int );
extern void display_prg_1(struct svc_req *rqstp, 
			  register SVCXPRT *transp);

/* The RPC Listener function */
void SVCService_Run() {
  fd_set  readfdset;
  extern int errno;
  struct timeval tv;
  static int  tsize = 0;

  tv.tv_sec  = 0;
  tv.tv_usec = 300000;

  if (!tsize) tsize = getdtablesize();

  while(1) {
    readfdset = svc_fdset;

    switch( select(tsize, &readfdset, (fd_set*)NULL, (fd_set*)NULL,
		   &tv )) {
    case -1:
      if (errno == EBADF) continue;
      perror("select failed\n");
      return;
    case 0:
      /* Collect zombies from fork when not active     */
      SVCCollectZombies(0);
      /* Also check for inserted hot pluggable devices */
      HPSearchHotPluggables();
      tv.tv_usec = 300000;
      continue;
    default:

      /* Only obey if there are less than allowed simultaneous connections
	 this will help against DOS attacks which might try to spawn off
	 many threads until the system comes down to a halting crash       */

      if ( SVCReturnChannelNumber() >= (PCSCLITE_MAX_CONTEXTS-1) ||
	   SVCReturnProcessNumber() >= (PCSCLITE_MAX_THREADS-1) ) {

	DebugLogA("SVCService_Run: Maximum channels/threads reached \n", 
		  __FILE__, __LINE__);
	
	/* Collect some zombies and sleep - let the OS take the attack */
	SVCCollectZombies(0);	
	SYS_Sleep(1);
      }
      
      svc_getreqset(&readfdset);
    }
  }
}

int main(int argc, char **argv) {
  
  register SVCXPRT *transp;
  transp = 0;

  /* Handle any command line arguments */
  if ( argc == 2 && (strcmp(argv[1], "-v") == 0) ) {
    printf("pcsc-lite version: %s <corcoran@linuxnet.com>\n",
           PCSCLITE_VERSION_NUMBER);
    return 0;
  } else if ( argc > 1 ) {
    printf("pcsc-lite: invalid arguments\n");
    return 1;
  }

  (void) pmap_unset(DISPLAY_PRG, DISPLAY_VER);
  
  transp = svctcp_create(RPC_ANYSOCK, 0, 0);
  if (transp == NULL) {
    fprintf(stderr, "cannot create tcp service.\n");
    return 1;
  }

  if (!svc_register(transp, DISPLAY_PRG, DISPLAY_VER, 
		    display_prg_1, IPPROTO_TCP)) {
    fprintf(stderr, "unable to register (DISPLAY_PRG, DISPLAY_VER, tcp).\n");
    return 1;
  }
  
  /* Allocate memory for reader structures     */
  RFAllocateReaderSpace( PCSCLITE_MAX_CONTEXTS );

  /* Grab the information from the reader.conf */    
  DBUpdateReaders ( PCSCLITE_READER_CONFIG );

  /* Set the default globals                   */
  g_rgSCardT0Pci.dwProtocol  = SCARD_PROTOCOL_T0;
  g_rgSCardT1Pci.dwProtocol  = SCARD_PROTOCOL_T1;  
  g_rgSCardRawPci.dwProtocol = SCARD_PROTOCOL_RAW;
  
  DebugLogA("main: PC/SC Lite Daemon Ready.\n", __FILE__, __LINE__);
  
  SVCService_Run();

  fprintf(stderr, "svc_run returned\n");
  return 1;
}
