/******************************************************************

	MUSCLE SmartCard Development ( http://www.linuxnet.com )
	    Title  : winscard.c
	    Package: pcsc lite
            Author : David Corcoran
            Date   : 10/27/99
	    License: Copyright (C) 1999 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This handles smartcard reader communications. 
	             This file forwards requests over RPC.

********************************************************************/

#include <winscard.h>
#include <string.h>
#include <sys_generic.h>
#include "winscard_rpc.h"

static struct _psChannelMap {
  SCARDCONTEXT hContext;
  SCARDHANDLE  hCard;
  SCARDHANDLE  hCardTable[PCSCLITE_MAX_CONTEXTS];
  DWORD        dwTableSize;
  CLIENT       *psChannel;
  LPSTR        psHost;
} *psChannelMap[PCSCLITE_MAX_CHANNELS];

static DWORD dwChannelNum = 0;

SCARD_IO_REQUEST g_rgSCardT0Pci, g_rgSCardT1Pci, g_rgSCardRawPci;   

/* These functions allow this file to map SCARDCONTEXT's and
   SCARDHANDLE's over to their appropriate CLIENT structure
   for RPC.  The Add and Remove functions allow the addition
   of multiple SCARDHANDLE's to one SCARDCONTEXT.             */

static LONG SCardAddChannel    ( SCARDCONTEXT );
static LONG SCardRemoveChannel ( SCARDCONTEXT );
static LONG SCardAddContext    ( SCARDCONTEXT, SCARDHANDLE );
static LONG SCardRemoveContext ( SCARDHANDLE );

LONG    SCardGetContext    ( SCARDCONTEXT, SCARDHANDLE );
CLIENT* SCardGetClient     ( LONG );

LONG SCardEstablishContext( DWORD dwScope, LPCVOID pvReserved1, 
			    LPCVOID pvReserved2, LPSCARDCONTEXT phContext) {

  LPSTR host;
  LONG liIndex, rv;
  CLIENT *clnt;
  establish_struct *result_1;
  establish_struct scardestablishcontext_1_arg;      

  /* Zero out everything   */
  host=0; liIndex=0; rv=0; clnt=0; result_1=0;
  memset( &scardestablishcontext_1_arg, 0x00, sizeof(establish_struct) );

  /* Get the services host */
  if ( pvReserved1 == 0 ) {
    host = strdup("localhost");
  } else {
    host = strdup((LPSTR)pvReserved1);
  }  

  *phContext = 0;

  g_rgSCardT0Pci.dwProtocol  = SCARD_PROTOCOL_T0;
  g_rgSCardT1Pci.dwProtocol  = SCARD_PROTOCOL_T1;
  g_rgSCardRawPci.dwProtocol = SCARD_PROTOCOL_RAW;

  clnt = clnt_create(host, DISPLAY_PRG, DISPLAY_VER, "tcp");
  
  if ( clnt != 0 ) {
    result_1 = scardestablishcontext_1(&scardestablishcontext_1_arg, clnt);
  } 

  if ( clnt == 0 || result_1 == 0 || result_1->rv != SCARD_S_SUCCESS ) {
    clnt_pcreateerror(host);  
    free ( host ); host = 0;
    return SCARD_E_NO_SERVICE;
  }

  rv = SCardAddChannel( result_1->phContext );

  if ( rv != SCARD_S_SUCCESS ) {
    free ( host ); host = 0;
    return rv;
  }

  liIndex = SCardGetContext( result_1->phContext, 0 );
  psChannelMap[liIndex]->psHost    = strdup(host);
  psChannelMap[liIndex]->psChannel = clnt;
  *phContext                       = result_1->phContext;

  SCardSetTimeout( *phContext, PCSCLITE_TRANSACTION_TIMEOUT );  

  free ( host ); host = 0;
  return SCARD_S_SUCCESS;
}

LONG SCardReleaseContext( SCARDCONTEXT hContext ) {

  LONG liIndex;
  CLIENT *clnt;
  release_struct  *result_2;
  release_struct  scardreleasecontext_1_arg;   

  /* Zero out everything   */
  liIndex=0; clnt=0; result_2=0;
  memset( &scardreleasecontext_1_arg, 0x00, sizeof(release_struct) );

  liIndex = SCardGetContext( hContext , 0 );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }

  scardreleasecontext_1_arg.hContext = hContext;

  clnt     = psChannelMap[liIndex]->psChannel;
  result_2 = scardreleasecontext_1(&scardreleasecontext_1_arg, clnt);

  SCardRemoveChannel( hContext );
  clnt_destroy( clnt );

  return SCARD_S_SUCCESS;
}

LONG SCardSetTimeout( SCARDCONTEXT hContext, DWORD dwTimeout ) {

  CLIENT *clnt;
  LONG liIndex;
  struct timeval rpctime;

  /* Zero out everything   */
  liIndex=0; clnt=0; 
  memset( &rpctime, 0x00, sizeof(struct timeval) );

  liIndex = SCardGetContext( hContext, 0 );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }

  rpctime.tv_sec  = dwTimeout;
  rpctime.tv_usec = 0;

  clnt     = psChannelMap[liIndex]->psChannel;  
  clnt_control( clnt, CLSET_TIMEOUT, (char *)&rpctime );
  
  return SCARD_S_SUCCESS;
}

LONG SCardConnect( SCARDCONTEXT hContext, LPCSTR szReader, DWORD dwShareMode,
                   DWORD dwPreferredProtocols, LPSCARDHANDLE phCard,
                   LPDWORD pdwActiveProtocol ) {              
  CLIENT *clnt;
  LONG liIndex;
  connect_struct  *result_3;
  connect_struct  scardconnect_1_arg;   

  /* Zero out everything   */
  liIndex=0; clnt=0; result_3=0;
  memset( &scardconnect_1_arg, 0x00, sizeof(connect_struct) );
  
  liIndex = SCardGetContext( hContext, 0 );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }

  scardconnect_1_arg.hContext                = hContext;
  scardconnect_1_arg.szReader.RPC_LPCSTR_val = strdup((LPSTR)szReader);
  scardconnect_1_arg.szReader.RPC_LPCSTR_len = strlen(szReader) + 1;
  scardconnect_1_arg.dwShareMode             = dwShareMode; 
  scardconnect_1_arg.dwPreferredProtocols    = dwPreferredProtocols;
  scardconnect_1_arg.phCard                  = *phCard;
  scardconnect_1_arg.pdwActiveProtocol       = *pdwActiveProtocol;

  clnt     = psChannelMap[liIndex]->psChannel;
  result_3 = scardconnect_1(&scardconnect_1_arg, clnt);  

  free(scardconnect_1_arg.szReader.RPC_LPCSTR_val);
  scardconnect_1_arg.szReader.RPC_LPCSTR_val = 0;  

  if ( result_3 == 0 ) {
    *phCard = 0;
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  }

  *phCard                                 = result_3->phCard;
  *pdwActiveProtocol                      = result_3->pdwActiveProtocol;


  if ( result_3->rv == SCARD_S_SUCCESS ) {
    SCardAddContext( hContext, result_3->phCard );
  }

  return result_3->rv;
}

LONG SCardReconnect( SCARDHANDLE hCard, DWORD dwShareMode,
		     DWORD dwPreferredProtocols, DWORD dwInitialization,
		     LPDWORD pdwActiveProtocol ) {              
  CLIENT *clnt;
  LONG liIndex;
  reconnect_struct  *result_3;
  reconnect_struct  scardreconnect_1_arg;   

  /* Zero out everything   */
  liIndex=0; clnt=0; result_3=0;
  memset( &scardreconnect_1_arg, 0x00, sizeof(reconnect_struct) );
  
  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }

  scardreconnect_1_arg.hCard                   = hCard;
  scardreconnect_1_arg.dwShareMode             = dwShareMode; 
  scardreconnect_1_arg.dwPreferredProtocols    = dwPreferredProtocols;
  scardreconnect_1_arg.dwInitialization        = dwInitialization;
  scardreconnect_1_arg.pdwActiveProtocol       = *pdwActiveProtocol;

  clnt     = psChannelMap[liIndex]->psChannel;
  result_3 = scardreconnect_1(&scardreconnect_1_arg, clnt);  

  if ( result_3 == 0 ) {
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  }

  *pdwActiveProtocol                      = result_3->pdwActiveProtocol;

  return result_3->rv;
}

LONG SCardDisconnect( SCARDHANDLE hCard, DWORD dwDisposition ) {
  
  CLIENT *clnt;
  LONG liIndex;
  disconnect_struct  *result_4;
  disconnect_struct  scarddisconnect_1_arg;

  /* Zero out everything   */
  liIndex=0; clnt=0; result_4=0;
  memset( &scarddisconnect_1_arg, 0x00, sizeof(disconnect_struct) );
  
  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }

  scarddisconnect_1_arg.hCard         = hCard;
  scarddisconnect_1_arg.dwDisposition = dwDisposition; 

  clnt     = psChannelMap[liIndex]->psChannel;
  result_4 = scarddisconnect_1(&scarddisconnect_1_arg, clnt);

  if ( result_4 == 0 ) {
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  }

  SCardRemoveContext( hCard );
  
  return result_4->rv;
}

LONG SCardBeginTransaction( SCARDHANDLE hCard ) {

  CLIENT *clnt;
  LONG liIndex;
  begin_struct  *result_5;
  begin_struct  scardbegintransaction_1_arg;  
  int timeval, randnum;
  
  /* Zero out everything   */
  liIndex=0; clnt=0; result_5=0; timeval=0; randnum=0;
  memset( &scardbegintransaction_1_arg, 0x00, sizeof(begin_struct) );

  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }
  
  scardbegintransaction_1_arg.hCard = hCard;
  clnt = psChannelMap[liIndex]->psChannel;
  
  /* Query the server every so often until the sharing violation
     ends and then hold the lock for yourself.                   */

  do {
    result_5 = scardbegintransaction_1(&scardbegintransaction_1_arg, clnt);
    
    if ( result_5 == 0 ) {
      clnt_perror(clnt, "call failed:"); 
      return SCARD_F_COMM_ERROR;
    }
    
    /* This helps prevent deadlocks */
    randnum = ( randnum, 100000.0, 300000.0 );
    SYS_USleep( randnum );

    timeval += 1;
  } while ( result_5->rv == SCARD_E_SHARING_VIOLATION &&
	    timeval < PCSCLITE_TRANSACTION_TIMEOUT );

  if ( timeval >= PCSCLITE_TRANSACTION_TIMEOUT ) {
    return SCARD_E_TIMEOUT;
  } else {
    return result_5->rv;
  }
}

LONG SCardEndTransaction( SCARDHANDLE hCard, DWORD dwDisposition ) {

  CLIENT *clnt;
  LONG liIndex;
  end_struct  *result_6;
  end_struct  scardendtransaction_1_arg;  
  int randnum;
  
  /* Zero out everything   */
  liIndex=0; clnt=0; result_6=0; randnum=0;
  memset( &scardendtransaction_1_arg, 0x00, sizeof(end_struct) );

  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }
  
  scardendtransaction_1_arg.hCard         = hCard;
  scardendtransaction_1_arg.dwDisposition = dwDisposition;

  clnt     = psChannelMap[liIndex]->psChannel;
  result_6 = scardendtransaction_1(&scardendtransaction_1_arg, clnt);

  if ( result_6 == 0 ) {
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  }

  /* This helps prevent deadlocks */
  randnum = ( randnum, 300000.0, 600000.0 );
  SYS_USleep( randnum );

  return result_6->rv;
}

LONG SCardCancelTransaction( SCARDHANDLE hCard ) {

  CLIENT *clnt;
  LONG liIndex;
  cancel_struct  *result_7;
  cancel_struct  scardcanceltransaction_1_arg;  
  
  /* Zero out everything   */
  liIndex=0; clnt=0; result_7=0;
  memset( &scardcanceltransaction_1_arg, 0x00, sizeof(cancel_struct) );

  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }
  
  scardcanceltransaction_1_arg.hCard = hCard;

  clnt     = psChannelMap[liIndex]->psChannel;
  result_7 = scardcanceltransaction_1(&scardcanceltransaction_1_arg, clnt);

  if ( result_7 == 0 ) {
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  }

  return result_7->rv;
}

LONG SCardStatus( SCARDHANDLE hCard, LPSTR mszReaderNames, 
		  LPDWORD pcchReaderLen, LPDWORD pdwState, 
		  LPDWORD pdwProtocol, LPBYTE pbAtr,
                  LPDWORD pcbAtrLen ) {
  
  LONG liIndex;
  CLIENT *clnt;
  status_struct *result_8;
  status_struct scardstatus_1_arg;

  /* Zero out everything   */
  liIndex=0; clnt=0; result_8=0;
  memset( &scardstatus_1_arg, 0x00, sizeof(status_struct) );

  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    *pcchReaderLen    = 0;
    *pcbAtrLen        = 0; 
    *pdwState         = 0;
    *pdwProtocol      = 0;
    return SCARD_E_INVALID_HANDLE;
  }

  scardstatus_1_arg.hCard                        = hCard;
  scardstatus_1_arg.mszReaderNames.RPC_LPSTR_val = 0;
  scardstatus_1_arg.mszReaderNames.RPC_LPSTR_len = 0; 
  scardstatus_1_arg.pbAtr.RPC_LPBYTE_val         = 0;
  scardstatus_1_arg.pbAtr.RPC_LPBYTE_len         = 0;
  if(mszReaderNames)
    scardstatus_1_arg.pcchReaderLen = *pcchReaderLen;
  else 
    scardstatus_1_arg.pcchReaderLen = RPC_MAX_STRING_SIZE;

  clnt     = psChannelMap[liIndex]->psChannel;
  result_8 = scardstatus_1(&scardstatus_1_arg, clnt);

  if ( result_8->rv == SCARD_S_SUCCESS ) {
    if(mszReaderNames) {
      memcpy( mszReaderNames, result_8->mszReaderNames.RPC_LPSTR_val,
              result_8->mszReaderNames.RPC_LPSTR_len );
    }
    memcpy( pbAtr, result_8->pbAtr.RPC_LPBYTE_val, 
	    result_8->pbAtr.RPC_LPBYTE_len );
  }

  *pcchReaderLen = result_8->pcchReaderLen;
  *pdwState      = result_8->pdwState;
  *pdwProtocol   = result_8->pdwProtocol;
  *pcbAtrLen     = result_8->pcbAtrLen;

  return result_8->rv;
}

LONG SCardGetStatusChange( SCARDCONTEXT hContext, DWORD dwTimeout, 
			   LPSCARD_READERSTATE_A rgReaderStates, 
                           DWORD cReaders ) {

  CLIENT *clnt;
  LONG liIndex;
  statuschange_struct  *result_9;
  statuschange_struct  scardgetstatuschange_1_arg;   

  /* Zero out everything   */
  liIndex=0; clnt=0; result_9=0;
  memset( &scardgetstatuschange_1_arg, 0x00, sizeof(statuschange_struct) );

  liIndex = SCardGetContext( hContext, 0 );

  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }
  
  scardgetstatuschange_1_arg.hContext  = hContext;
  scardgetstatuschange_1_arg.dwTimeout = dwTimeout;
  scardgetstatuschange_1_arg.cReaders  = cReaders;

  scardgetstatuschange_1_arg.rgReaderStates.szReader.RPC_LPCSTR_val = 
    strdup((LPSTR)rgReaderStates->szReader);
  scardgetstatuschange_1_arg.rgReaderStates.szReader.RPC_LPCSTR_len =
    strlen(rgReaderStates->szReader) + 1;
  scardgetstatuschange_1_arg.rgReaderStates.dwCurrentState = 
    rgReaderStates->dwCurrentState;
  scardgetstatuschange_1_arg.rgReaderStates.dwEventState = 
    rgReaderStates->dwEventState;
  scardgetstatuschange_1_arg.rgReaderStates.cbAtr                 = 0;
  scardgetstatuschange_1_arg.rgReaderStates.rgbAtr.RPC_LPBYTE_len = 0;;

  if ( dwTimeout == INFINITE ) {
    /* Infinite in our case is 50 days */
    SCardSetTimeout( hContext, PCSCLITE_INFINITE_TIMEOUT );
  }

  clnt     = psChannelMap[liIndex]->psChannel;
  result_9 = scardgetstatuschange_1(&scardgetstatuschange_1_arg, clnt);
  
  SCardSetTimeout( hContext, PCSCLITE_TRANSACTION_TIMEOUT );
  
  if ( result_9 == 0 ) {
    free(scardgetstatuschange_1_arg.rgReaderStates.szReader.RPC_LPCSTR_val);
    scardgetstatuschange_1_arg.rgReaderStates.szReader.RPC_LPCSTR_val = 0;
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  }

  rgReaderStates->dwCurrentState = result_9->rgReaderStates.dwCurrentState;
  rgReaderStates->dwEventState   = result_9->rgReaderStates.dwEventState;
  rgReaderStates->cbAtr          = result_9->rgReaderStates.cbAtr; 

  memcpy( rgReaderStates->rgbAtr, 
	  result_9->rgReaderStates.rgbAtr.RPC_LPBYTE_val, 
	  result_9->rgReaderStates.rgbAtr.RPC_LPBYTE_len );

  free(scardgetstatuschange_1_arg.rgReaderStates.szReader.RPC_LPCSTR_val);
  scardgetstatuschange_1_arg.rgReaderStates.szReader.RPC_LPCSTR_val = 0;

  return result_9->rv;
}

LONG SCardControl( SCARDHANDLE hCard, LPCBYTE pbSendBuffer, DWORD cbSendLength,
                   LPBYTE pbRecvBuffer, LPDWORD pcbRecvLength ) {

  SCARD_IO_REQUEST pioSendPci, pioRecvPci;

  pioSendPci.dwProtocol = SCARD_PROTOCOL_RAW;
  pioRecvPci.dwProtocol = SCARD_PROTOCOL_RAW;

  return SCardTransmit( hCard, &pioSendPci, pbSendBuffer, cbSendLength, 
                        &pioRecvPci, pbRecvBuffer, pcbRecvLength );
}

LONG SCardTransmit( SCARDHANDLE hCard, LPCSCARD_IO_REQUEST pioSendPci,
                    LPCBYTE pbSendBuffer, DWORD cbSendLength,
                    LPSCARD_IO_REQUEST pioRecvPci, LPBYTE pbRecvBuffer, 
		    LPDWORD pcbRecvLength ) {
 
  CLIENT *clnt;
  LONG liIndex;
  transmit_struct  *result_10;
  transmit_struct  scardtransmit_1_arg;  
  
  /* Zero out everything   */
  liIndex=0; clnt=0; result_10=0;
  memset( &scardtransmit_1_arg, 0x00, sizeof(transmit_struct) );

  liIndex = SCardGetContext( 0, hCard );

  if ( liIndex < 0 ) {
    *pcbRecvLength = 0;
    return SCARD_E_INVALID_HANDLE;
  }

  scardtransmit_1_arg.hCard                        = hCard;
  scardtransmit_1_arg.cbSendLength                 = cbSendLength;
  scardtransmit_1_arg.pcbRecvLength                = *pcbRecvLength;
  scardtransmit_1_arg.pbSendBuffer.RPC_LPCBYTE_len = cbSendLength;
  scardtransmit_1_arg.pbRecvBuffer.RPC_LPBYTE_len  = 0;

  scardtransmit_1_arg.pbSendBuffer.RPC_LPCBYTE_val = (char *)malloc
    (sizeof(char)*cbSendLength);

  memcpy( &scardtransmit_1_arg.pioSendPci, pioSendPci, 
	  sizeof(SCARD_IO_REQUEST) );
  
  memcpy( scardtransmit_1_arg.pbSendBuffer.RPC_LPCBYTE_val, 
	  pbSendBuffer, cbSendLength );

  clnt      = psChannelMap[liIndex]->psChannel;
  result_10 = scardtransmit_1(&scardtransmit_1_arg, clnt);
  

  /* Zero it and free it so any secret information cannot be leaked */
  memset( scardtransmit_1_arg.pbSendBuffer.RPC_LPCBYTE_val, 
	  0x00, cbSendLength );
  free ( scardtransmit_1_arg.pbSendBuffer.RPC_LPCBYTE_val );
  scardtransmit_1_arg.pbSendBuffer.RPC_LPCBYTE_val = 0;
  
  if ( result_10 == 0 ) {
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  } 

  *pcbRecvLength = result_10->pcbRecvLength;

  /* Copy and zero it so any secret information is not leaked       */
  memcpy( pbRecvBuffer, result_10->pbRecvBuffer.RPC_LPBYTE_val, 
	  result_10->pcbRecvLength );
  memset( result_10->pbRecvBuffer.RPC_LPBYTE_val, 
	  0x00, result_10->pcbRecvLength );

  if ( pioRecvPci ) {
    memcpy( pioRecvPci, &result_10->pioRecvPci, 
	    sizeof(SCARD_IO_REQUEST) );
  }

  return result_10->rv;
}

LONG SCardListReaders( SCARDCONTEXT hContext, LPCSTR mszGroups, 
		       LPSTR mszReaders, LPDWORD pcchReaders ) {

  CLIENT *clnt;
  LONG liIndex;
  LPSTR mszGroupsB;
  DWORD dwGroupsLen, dwReadersLen;
  list_struct  *result_11;
  list_struct  scardlistreaders_1_arg; 
  
  /* Zero out everything   */
  liIndex=0; clnt=0; mszGroupsB=0; dwGroupsLen=0; dwReadersLen=0; result_11=0;
  memset( &scardlistreaders_1_arg, 0x00, sizeof(list_struct) );

  liIndex = SCardGetContext( hContext, 0 );
  
  if ( liIndex < 0 ) {
    *pcchReaders = 0;
    return SCARD_E_INVALID_HANDLE;
  }

  /* Groups are not currently used */
  mszGroupsB  = 0;
  dwGroupsLen = 0;  

  if ( mszReaders == 0 ) {
    dwReadersLen = 0;
  } else {
    /*    dwReadersLen = strlen(mszReaders)+1;  */
    dwReadersLen = *pcchReaders;
  }

  scardlistreaders_1_arg.hContext                  = hContext;
  scardlistreaders_1_arg.mszGroups.RPC_LPCSTR_val  = (LPSTR)mszGroupsB;
  scardlistreaders_1_arg.mszGroups.RPC_LPCSTR_len  = dwGroupsLen;
  scardlistreaders_1_arg.mszReaders.RPC_LPSTR_val  = mszReaders;
  scardlistreaders_1_arg.mszReaders.RPC_LPSTR_len  = dwReadersLen; 
  scardlistreaders_1_arg.pcchReaders               = *pcchReaders;

  clnt       = psChannelMap[liIndex]->psChannel;
  result_11  = scardlistreaders_1(&scardlistreaders_1_arg, clnt); 

  if ( result_11 == 0 ) {
    clnt_perror(clnt, "call failed:");
    *pcchReaders = 0;
    return SCARD_F_COMM_ERROR;
  } 

  if ( mszReaders != 0 ) {
    memcpy( mszReaders, result_11->mszReaders.RPC_LPSTR_val, 
	    *pcchReaders );
  }

  *pcchReaders = result_11->pcchReaders;
  return result_11->rv;
}


/* For compatibility purposes only */
LONG SCardListReaderGroups( SCARDCONTEXT hContext, LPSTR mszGroups, 
                            LPDWORD pcchGroups)
{

  LONG rv = SCARD_S_SUCCESS;

  const char ReaderGroup[] = "SCard$DefaultReaders";
  const dwGroups = strlen(ReaderGroup)+2;
    
  if(mszGroups) {
    
    if(*pcchGroups < dwGroups)
      rv = SCARD_E_INSUFFICIENT_BUFFER;
    else {
      memset(mszGroups,0,dwGroups);
      memcpy(mszGroups,ReaderGroup,strlen(ReaderGroup));
    }
  }

  *pcchGroups = dwGroups;
  
  return rv;
}

LONG SCardCancel( SCARDCONTEXT hContext ) {

  CLIENT *clnt;
  LONG liIndex;
  cancellock_struct  *result_12;
  cancellock_struct  scardcancel_1_arg;  
 
  /* Zero out everything   */
  liIndex=0; clnt=0; result_12=0;
  memset( &scardcancel_1_arg, 0x00, sizeof(cancellock_struct) );

  liIndex = SCardGetContext( hContext, 0 );
  
  if ( liIndex < 0 ) {
    return SCARD_E_INVALID_HANDLE;
  }

  scardcancel_1_arg.hContext = hContext;

  clnt      = psChannelMap[liIndex]->psChannel;
  result_12 = scardcancel_1(&scardcancel_1_arg, clnt);  
  
  if ( result_12 == 0 ) {
    clnt_perror(clnt, "call failed:"); 
    return SCARD_F_COMM_ERROR;
  } 

  return result_12->rv;
}

LONG SCardGetContext( SCARDCONTEXT hContext, SCARDHANDLE hCard ) {

  int i, p;
  i=0; p=0;

  if ( hCard == 0 ) {
    for ( i=0; i<dwChannelNum; i++ ) {
      if ( hContext == psChannelMap[i]->hContext 
	   && psChannelMap[i] != 0 ) {
	return i;
      }
    }
  } else {
    for ( i=0; i<dwChannelNum; i++ ) {
      for (p=0; p < psChannelMap[i]->dwTableSize; p++) {
	if ( hCard == psChannelMap[i]->hCardTable[p] ) {
	  return i;
	}
      }
    }
  }    
  
  return -1;
}

CLIENT* SCardGetClient( LONG liIndex ) {
  return psChannelMap[liIndex]->psChannel;
}

LONG SCardAddChannel( SCARDCONTEXT hContext ) {
  
  int i;
  i=0;

  if ( dwChannelNum >= PCSCLITE_MAX_CHANNELS ) {
    return SCARD_E_INSUFFICIENT_BUFFER;
  }
  
  /* See if channel already exists */

  for ( i=0; i < dwChannelNum; i++ ) {
    if ( hContext == psChannelMap[i]->hContext ) {
      return SCARD_S_SUCCESS;
    }
  }

  psChannelMap[dwChannelNum] = (struct _psChannelMap *)calloc
    (1,sizeof(struct _psChannelMap));
  
  psChannelMap[dwChannelNum]->hContext    = hContext;
  psChannelMap[dwChannelNum]->dwTableSize = 0;
  psChannelMap[dwChannelNum]->psChannel   = 0;
  psChannelMap[dwChannelNum]->psHost      = 0;

  dwChannelNum += 1;

  return SCARD_S_SUCCESS;
}

LONG SCardRemoveChannel( SCARDCONTEXT hContext ) {
  
  LONG liIndex;
  liIndex=0;

  liIndex = SCardGetContext( hContext, 0 );
  
  if ( liIndex != -1 ) {
    
    /* The server will do the disconnects
       and cancel requests.               */
    
    if ( psChannelMap[liIndex]->psHost != 0 ) {
      free( psChannelMap[liIndex]->psHost );
      psChannelMap[liIndex]->psHost = 0;
    }

    free( psChannelMap[liIndex] );
    psChannelMap[liIndex] = 0;

    dwChannelNum          -= 1;
    psChannelMap[liIndex]  = psChannelMap[dwChannelNum];
    return SCARD_S_SUCCESS;
  } 
  
  return SCARD_E_INVALID_HANDLE;
}

LONG SCardAddContext( SCARDCONTEXT hContext, SCARDHANDLE hCard ) {

  /* Add a SCARDHANDLE to the current Channel
     Increment the Table Size                 */


  LONG liIndex;
  int i;

  liIndex=0; i=0;
  
  liIndex = SCardGetContext( hContext, 0 );
  
  if ( liIndex != -1 ) {

    /* See if context already exists */

    for ( i=0; i < psChannelMap[liIndex]->dwTableSize; i++ ) {
      if ( hCard == psChannelMap[liIndex]->hCardTable[i] ) {
	return SCARD_S_SUCCESS;
      }
    }

    if ( psChannelMap[liIndex]->dwTableSize >= PCSCLITE_MAX_CONTEXTS ) {
      return SCARD_E_INSUFFICIENT_BUFFER;
    }
    
    psChannelMap[liIndex]->hCardTable[psChannelMap[liIndex]->
				     dwTableSize] = hCard;
    psChannelMap[liIndex]->dwTableSize += 1;
    return SCARD_S_SUCCESS;
  }

  return SCARD_E_INVALID_HANDLE;
}

LONG SCardRemoveContext ( SCARDHANDLE hCard ) {

  int i,p;

  i=0; p=0;

  /* Lookup the hCard in all the Context Structures
     Delete it and decrement the counter.
     Note: There is a very low probability that two
     Channels might have the same hCard value in
     which this might fail.  FIX : There is a 1:65536
     chance of this happenning only if connections
     are made to more than 1 resource manager        */

  for (i=0; i < dwChannelNum; i++) {
    for (p=0; p < psChannelMap[i]->dwTableSize; p++) {
      if ( psChannelMap[i]->hCardTable[p] == hCard ) {
	psChannelMap[i]->dwTableSize   -= 1;
	psChannelMap[i]->hCardTable[p]  = 
	  psChannelMap[i]->hCardTable[psChannelMap[i]->dwTableSize];
      }
    }
  }

  return SCARD_S_SUCCESS;
}
