/******************************************************************

            Title  : winscard_rpc_server.c
            Package: PC/SC Lite
            Author : David Corcoran
            Date   : 09/17/99
	    License: Copyright (C) 1999 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This handles server related rpc calls to the
                     resource manager.

********************************************************************/

#include "winscard_rpc.h"
#include <winscard.h>
#include <debuglog.h>
#include <sys_generic.h>
#include <string.h>

extern int errno;

static struct _psChannelMap {
  SCARDCONTEXT hContext;
  SCARDCONTEXT hStaticContext;
  SCARDHANDLE  hCardTable[PCSCLITE_MAX_CONTEXTS];
  DWORD        dwTableSize;
  DWORD        psClient;
  CLIENT       *psChannel;
  LPSTR        psHost;
} psChannelMap[PCSCLITE_MAX_CHANNELS];

static DWORD dwChannelNum = 0;

/* Some local functions    */

static LONG SVCAddChannel    ( SCARDCONTEXT );
static LONG SVCRemoveChannel ( SCARDCONTEXT );
static LONG SVCGetContext    ( SCARDCONTEXT, SCARDHANDLE );
static LONG SVCAddContext    ( SCARDCONTEXT, SCARDHANDLE );
static LONG SVCRemoveContext ( SCARDHANDLE );

/* Some exported functions */

void SVCCollectZombies( int );
int  SVCReturnChannelNumber( void );

/* Some exported variables */

extern DWORD dwProcessNum;

establish_struct * 
scardestablishcontext_1_svc(establish_struct *argp, struct svc_req *rqstp)
{
  LONG liIndex, rv;
  SCARDCONTEXT hContext;
  static establish_struct  result;

  /* Zero out everything */
  liIndex=0; rv=0; hContext=0;

  /* Lets clean up and see if their are any 
     zombie connections still in existence   */

  SVCCollectZombies(rqstp->rq_xprt->xp_sock);
  rv = SCardEstablishContext( argp->dwScope, 0, 0, &argp->phContext );
  result.phContext = argp->phContext;
  
  if ( rv == SCARD_S_SUCCESS ) {
    hContext = argp->phContext;
    rv       = SVCAddChannel( hContext );
    
    if ( rv == SCARD_S_SUCCESS ) {
      liIndex = SVCGetContext( hContext, 0 );
      psChannelMap[liIndex].psClient       = rqstp->rq_xprt->xp_sock;
      psChannelMap[liIndex].hStaticContext = argp->phContext; 

#ifdef PCSC_DEBUG
      DebugLogB("scardestablishcontext_1_svc: Current Channel: %x\n",
		hContext, __FILE__, __LINE__ );
      DebugLogB("scardestablishcontext_1_svc: Current Index  : %d\n",
		liIndex, __FILE__, __LINE__ );      
#endif

    } else {
      argp->phContext = 0;
    }
  }

  result.rv = rv;
  return(&result);
}

release_struct * 
scardreleasecontext_1_svc(release_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  SCARDCONTEXT hContext;
  static release_struct  result;

  /* Zero out everything */
  rv=0; hContext=0;

  hContext = argp->hContext;
  rv = SVCRemoveChannel( hContext );
  
  if ( rv == SCARD_S_SUCCESS ) {
#ifdef PCSC_DEBUG
    DebugLogB("scardreleasecontext_1_svc: Release Channel: %x\n",
	      hContext, __FILE__, __LINE__ );
#endif
    
    rv = SCardReleaseContext( argp->hContext );
  }

  result.rv = rv;  
  return(&result);
}

connect_struct * 
scardconnect_1_svc(connect_struct *argp, struct svc_req *rqstp)
{
  LONG rv, rvb;
  SCARDCONTEXT hContext;
  static connect_struct  result;

  /* Zero out everything */
  rv=0; rvb=0; hContext=0;

  hContext = argp->hContext; 
  argp->phCard = 0;

  rv = SCardConnect( argp->hContext, argp->szReader.RPC_LPCSTR_val, 
		     argp->dwShareMode, argp->dwPreferredProtocols, 
		     &argp->phCard, &argp->pdwActiveProtocol );

  result.phCard            = argp->phCard;
  result.pdwActiveProtocol = argp->pdwActiveProtocol;

  /* Add the hCard to the list */
  if ( rv == SCARD_S_SUCCESS ) {
    rvb = SVCAddContext( hContext, result.phCard );
    if ( rvb != SCARD_S_SUCCESS ) {
      SCardDisconnect( argp->phCard, SCARD_LEAVE_CARD );
      argp->phCard = 0;
      rv = rvb;
    }
  }

  result.rv = rv;
  return(&result);
}

reconnect_struct * 
scardreconnect_1_svc(reconnect_struct *argp, struct svc_req *rqstp)
{
  LONG rv, rvb;
  static reconnect_struct  result;

  /* Zero out everything */
  rv=0; rvb=0;

  rv = SCardReconnect( argp->hCard, argp->dwShareMode, 
		       argp->dwPreferredProtocols, argp->dwInitialization, 
		       &argp->pdwActiveProtocol );

  result.pdwActiveProtocol = argp->pdwActiveProtocol;

  result.rv = rv;
  return(&result);
}

disconnect_struct * 
scarddisconnect_1_svc(disconnect_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static disconnect_struct  result;

  /* Zero out everything */
  rv=0; 

  /* Remove the hCard from the list */\
  rv = SVCRemoveContext( argp->hCard );

  if ( rv == SCARD_S_SUCCESS ) {
    rv = SCardDisconnect( argp->hCard, argp->dwDisposition );
  }

  result.rv = rv;
  return(&result);
}

begin_struct * 
scardbegintransaction_1_svc(begin_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static begin_struct  result;

  /* Zero out everything */
  rv=0; 

  rv = SCardBeginTransaction( argp->hCard );
  result.rv = rv;
  return(&result);
}

end_struct * 
scardendtransaction_1_svc(end_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static end_struct  result;

  /* Zero out everything */
  rv=0; 

  rv = SCardEndTransaction( argp->hCard, argp->dwDisposition );
  result.rv = rv;
  return(&result);
}

cancel_struct * 
scardcanceltransaction_1_svc(cancel_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static cancel_struct  result;

  /* Zero out everything */
  rv=0; 

  rv = SCardCancel( argp->hCard );
  result.rv = rv;
  return(&result);
}

status_struct * 
scardstatus_1_svc(status_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static status_struct  result;
  BYTE pbAtr[MAX_ATR_SIZE];
  LPSTR mszReaders;

  /* Zero out everything */
  rv=0; mszReaders=0; memset( pbAtr, 0x00, MAX_ATR_SIZE );

  mszReaders = (LPSTR)malloc(sizeof(char)*argp->pcchReaderLen);

  rv = SCardStatus( argp->hCard, mszReaders,
		    &argp->pcchReaderLen, &argp->pdwState,
		    &argp->pdwProtocol, pbAtr,
		    &argp->pcbAtrLen );

  if ( rv == SCARD_S_SUCCESS ) {

    /* Free up previously allocated memory */
    if ( result.pbAtr.RPC_LPBYTE_val != 0 ) {
      free ( result.pbAtr.RPC_LPBYTE_val );
      result.pbAtr.RPC_LPBYTE_val = 0;
      result.pbAtr.RPC_LPBYTE_len = 0;
    }

    if ( result.mszReaderNames.RPC_LPSTR_val != 0 ) {
      free ( result.mszReaderNames.RPC_LPSTR_val );
      result.mszReaderNames.RPC_LPSTR_val = 0;
      result.mszReaderNames.RPC_LPSTR_len = 0;	
    }

    /* Allocate some memory */
    result.pbAtr.RPC_LPBYTE_len         = argp->pcbAtrLen;
    result.pbAtr.RPC_LPBYTE_val         = 
      (LPBYTE)malloc(sizeof(BYTE)*argp->pcbAtrLen);
    result.mszReaderNames.RPC_LPSTR_len = argp->pcchReaderLen; 
    result.mszReaderNames.RPC_LPSTR_val = strdup(mszReaders);

    /* Copy the memory      */
    memcpy( result.pbAtr.RPC_LPBYTE_val, pbAtr, argp->pcbAtrLen );
  }

  result.pcchReaderLen                 = argp->pcchReaderLen;
  result.pdwState                      = argp->pdwState;
  result.pdwProtocol                   = argp->pdwProtocol;
  result.pcbAtrLen                     = argp->pcbAtrLen;
  result.rv                            = rv;

  free (mszReaders); mszReaders = 0;

  return(&result);
}

statuschange_struct * 
scardgetstatuschange_1_svc(statuschange_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static statuschange_struct  result;
  SCARD_READERSTATE_A sTempState;

  /* Zero out everything */
  rv=0; memset( &sTempState, 0x00, sizeof(SCARD_READERSTATE_A) );

  /* RPC does not like to pass voids so I have to remove the void
     and pass it along in a modified structure and then recopy it */

  sTempState.szReader       = argp->rgReaderStates.szReader.RPC_LPCSTR_val;
  sTempState.dwCurrentState = argp->rgReaderStates.dwCurrentState;
  sTempState.dwEventState   = argp->rgReaderStates.dwEventState;
  
  rv = SCardGetStatusChange( argp->hContext, argp->dwTimeout,
			     &sTempState, argp->cReaders );
  
  if ( result.rgReaderStates.szReader.RPC_LPCSTR_val != 0 ) {
    free ( result.rgReaderStates.szReader.RPC_LPCSTR_val );
    result.rgReaderStates.szReader.RPC_LPCSTR_val = 0;
    result.rgReaderStates.szReader.RPC_LPCSTR_len = 0;  
  }
  
  result.rgReaderStates.szReader.RPC_LPCSTR_val = strdup(sTempState.szReader);
  result.rgReaderStates.szReader.RPC_LPCSTR_len = strlen(sTempState.szReader)+1;
  result.rgReaderStates.dwCurrentState          = sTempState.dwCurrentState;
  result.rgReaderStates.dwEventState            = sTempState.dwEventState;
  result.rgReaderStates.cbAtr                   = sTempState.cbAtr;

  if ( sTempState.cbAtr <= MAX_ATR_SIZE ) {

    /* Free up any previously allocated memory */
    if ( result.rgReaderStates.rgbAtr.RPC_LPBYTE_val != 0 ) {
      free ( result.rgReaderStates.rgbAtr.RPC_LPBYTE_val );
      result.rgReaderStates.rgbAtr.RPC_LPBYTE_val = 0;
      result.rgReaderStates.rgbAtr.RPC_LPBYTE_len = 0;
    }

    result.rgReaderStates.rgbAtr.RPC_LPBYTE_len = sTempState.cbAtr;
    result.rgReaderStates.rgbAtr.RPC_LPBYTE_val = (BYTE *)malloc
      (sizeof(BYTE)*sTempState.cbAtr);

    memcpy( result.rgReaderStates.rgbAtr.RPC_LPBYTE_val, sTempState.rgbAtr, 
	    sTempState.cbAtr ); 
  } else {
    result.rgReaderStates.cbAtr = 0;
  }   

  result.rv = rv;
  return(&result);
}

transmit_struct * 
scardtransmit_1_svc(transmit_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static transmit_struct  result;
  SCARD_IO_REQUEST pioSendPci;
  SCARD_IO_REQUEST pioRecvPci;
  BYTE pbRecvBuffer[RPC_MAX_BUFFER_SIZE];
  
  /* Zero out everything */
  rv=0; memset( &pioSendPci, 0x00, sizeof(SCARD_IO_REQUEST) );
  memset( &pioRecvPci, 0x00, sizeof(SCARD_IO_REQUEST) );
  memset( pbRecvBuffer, 0x00, RPC_MAX_BUFFER_SIZE );

  pioSendPci.dwProtocol  = argp->pioSendPci.dwProtocol;
  pioSendPci.cbPciLength = argp->pioSendPci.cbPciLength;

  rv = SCardTransmit( argp->hCard, &pioSendPci, 
 		      argp->pbSendBuffer.RPC_LPCBYTE_val,
		      argp->cbSendLength, &pioRecvPci, 
		      pbRecvBuffer, 
		      &argp->pcbRecvLength );

  /* Free up any previously allocated memory */
  if ( result.pbRecvBuffer.RPC_LPBYTE_val != 0 ) {
    free (result.pbRecvBuffer.RPC_LPBYTE_val);
    result.pbRecvBuffer.RPC_LPBYTE_val = 0;
    result.pbRecvBuffer.RPC_LPBYTE_len = 0;
  }

  result.pcbRecvLength               = argp->pcbRecvLength; 
  result.pbRecvBuffer.RPC_LPBYTE_len = result.pcbRecvLength;
  result.pbRecvBuffer.RPC_LPBYTE_val = (char *)malloc
    (sizeof(char)*result.pcbRecvLength);

  /* Copy and erase - to keep secret information contained */
  memcpy( result.pbRecvBuffer.RPC_LPBYTE_val, pbRecvBuffer, 
	  result.pcbRecvLength );
  memset( pbRecvBuffer, 0x00, RPC_MAX_BUFFER_SIZE );

  memcpy( &result.pioRecvPci, &pioRecvPci, 
	  sizeof(SCARD_IO_REQUEST) );
  
  result.rv = rv;
  return(&result);
}

list_struct * 
scardlistreaders_1_svc(list_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static list_struct result;
  
  /* Zero out everything */
  rv=0; 

  if ( argp->mszReaders.RPC_LPSTR_val != 0 ) {
    if ( result.mszReaders.RPC_LPSTR_val != 0 ) {
      free (result.mszReaders.RPC_LPSTR_val);
      result.mszReaders.RPC_LPSTR_val = 0;
      result.mszReaders.RPC_LPSTR_len = 0;
    }

    result.mszReaders.RPC_LPSTR_len = argp->pcchReaders;
    result.mszReaders.RPC_LPSTR_val = 
      (char *)malloc(sizeof(char)*argp->pcchReaders);

    rv = SCardListReaders( argp->hContext, argp->mszGroups.RPC_LPCSTR_val,
			   result.mszReaders.RPC_LPSTR_val, &result.pcchReaders );

  } else {
    rv = SCardListReaders( argp->hContext, argp->mszGroups.RPC_LPCSTR_val,
			   0, &result.pcchReaders );
  }

  result.rv = rv;
  return(&result);
}

cancellock_struct *
scardcancel_1_svc(cancellock_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static cancellock_struct  result;
  rv = SCardCancel( argp->hContext );
  result.rv = rv;
  return(&result);
}                                                    

intro_struct *
scardintroducereader_1_svc(intro_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static intro_struct  result;

  rv = RFAddReader( argp->szReaderName.RPC_LPSTR_val,
		    argp->dwChannelID,
		    argp->szLibraryName.RPC_LPSTR_val );

  result.rv = rv;
  return(&result);
}                                                    

forget_struct *
scardforgetreader_1_svc(forget_struct *argp, struct svc_req *rqstp)
{
  LONG rv;
  static forget_struct  result;

  rv = RFRemoveReader( argp->szReaderName.RPC_LPSTR_val );		    

  result.rv = rv;
  return(&result);
}                                                    

                                                    
LONG SVCGetContext( SCARDCONTEXT hContext, SCARDHANDLE hCard ) {

  int i, p;

  if ( hCard == 0 ) {
    for ( i=0; i<dwChannelNum; i++ ) {
      if ( hContext == psChannelMap[i].hContext ) {
	return i;
      }
    }
  } else {
    for ( i=0; i<dwChannelNum; i++ ) {
      for (p=0; p < psChannelMap[i].dwTableSize; p++) {
	if ( hCard == psChannelMap[i].hCardTable[p] ) {
	  return i;
	}
      }
    }
  }    
  
  return -1;
}

LONG SVCAddChannel( SCARDCONTEXT hContext ) {

  int i;

  if ( dwChannelNum >= PCSCLITE_MAX_CHANNELS ) {
    return SCARD_E_INSUFFICIENT_BUFFER;
  }
  
  /* Check for duplicates */
  for ( i=0; i < PCSCLITE_MAX_CHANNELS; i++ ) {
    if ( hContext == psChannelMap[i].hContext ) {
      return SCARD_E_INVALID_HANDLE;
    }
  }

  psChannelMap[dwChannelNum].hContext    = hContext;
  psChannelMap[dwChannelNum].dwTableSize = 0;
  psChannelMap[dwChannelNum].psClient    = 0;
  psChannelMap[dwChannelNum].psChannel   = 0;
  psChannelMap[dwChannelNum].psHost      = 0;

  dwChannelNum += 1;

  return SCARD_S_SUCCESS;
}

LONG SVCRemoveChannel( SCARDCONTEXT hContext ) {
  
  LONG rv, liIndex;
  int i;
  
  liIndex = SVCGetContext( hContext, 0 );
  
  if ( liIndex != -1 ) {
    
    rv = SCardCancel( hContext );

    for ( i=0; i < psChannelMap[liIndex].dwTableSize; i++ ) {
      rv = SCardDisconnect( psChannelMap[liIndex].hCardTable[i], 
			    SCARD_UNPOWER_CARD );      
    }
    
    dwChannelNum          -= 1;
    psChannelMap[liIndex]  = psChannelMap[dwChannelNum];
    return SCARD_S_SUCCESS;
  } 
  
  return SCARD_E_INVALID_HANDLE;
}

LONG SVCAddContext( SCARDCONTEXT hContext, SCARDHANDLE hCard ) {

  /* Add a SCARDHANDLE to the current Channel
     Increment the Table Size                 */


  LONG liIndex;
  
  liIndex = SVCGetContext( hContext, 0 );
  
  if ( liIndex != -1 ) {
    if ( psChannelMap[liIndex].dwTableSize >= PCSCLITE_MAX_CONTEXTS ) {
      return SCARD_E_INSUFFICIENT_BUFFER;
    }
    
    psChannelMap[liIndex].hCardTable[psChannelMap[liIndex].
				     dwTableSize] = hCard;
    psChannelMap[liIndex].dwTableSize += 1;
    return SCARD_S_SUCCESS;
  }

  return SCARD_E_INVALID_HANDLE;
}

LONG SVCRemoveContext ( SCARDHANDLE hCard ) {

  int i,p;

  /* Lookup the hCard in all the Context Structures
     Delete it and decrement the counter.
     Note: There is a very low probability that two
     Channels might have the same hCard value in
     which this might fail.  FIX : There is a 1:65536
     chance of this happenning only if connections
     are made to more than 1 resource manager        */

  for (i=0; i < dwChannelNum; i++) {
    for (p=0; p < psChannelMap[i].dwTableSize; p++) {
      if ( psChannelMap[i].hCardTable[p] == hCard ) {
	psChannelMap[i].dwTableSize -= 1;
	psChannelMap[i].hCardTable[p] = 
	  psChannelMap[i].hCardTable[psChannelMap[i].dwTableSize];
      }
    }
  }

  return SCARD_S_SUCCESS;
}

int SVCReturnChannelNumber ( void ) {
  return dwChannelNum;
}

void SVCCollectZombies( int portfd ) {

  int i, ret, pid;
  DWORD psClient, dwChannels;

  /* Clean up zombie processes from 
     SCardGetStatusChange           */

  pid = SYS_Wait( 0, 0 ); 

  /* FIX :: A quick fix to determine whether processes have ended */
  if ( pid > 0 ) {
    dwProcessNum -= 1;
  }
  /* End of FIX ***************************************************/

  /* This is kind of strange in the meantime
     If a client dies the operating system will
     free the file descriptor associated with that
     socket connection so I check to see if the current
     descriptor matches a previous and remove any previous
  */

  dwChannels = dwChannelNum;
  for (i=0; i < dwChannels; i++) {
    psClient = psChannelMap[i].psClient;
    if ( psClient == portfd ) {
      if ( portfd != 0 ) {
	DebugLogB("SVCCollectZombies: Client %d Died\n", i,
		  __FILE__, __LINE__);     
	SVCRemoveChannel( psChannelMap[i].hContext );
      }
    } else {
      ret = SYS_Stat( psClient );
      if ( ret == -1 ) {
	DebugLogB("SVCCollectZombies: Client %d Died\n", i,
		  __FILE__, __LINE__);
	SVCRemoveChannel( psChannelMap[i].hContext );    
      }
    }
  }    
}
