/******************************************************************

            Title  : winscard_rpc_svc.c
            Package: PC/SC Lite RPC
            Author : David Corcoran
            Date   : 10/24/99
	    License: Copyright (C) 1999 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This handles the RPC server services.

********************************************************************/


#include "winscard_rpc.h"
#include "rpc_sechandler.h"
#include <sys_generic.h>
#include <debuglog.h>
#include <stdio.h>

#ifdef __STDC__
#define SIG_PF void(*)(int)
#endif

/* Some local definitions    */
DWORD dwProcessNum = 0;

/* Some exported definitions */
int SVCReturnProcessNumber( void );

void display_prg_1(struct svc_req *rqstp, register SVCXPRT *transp)
{
  union {
    establish_struct scardestablishcontext_1_arg;
    release_struct scardreleasecontext_1_arg;
    connect_struct scardconnect_1_arg;
    reconnect_struct scardreconnect_1_arg;
    disconnect_struct scarddisconnect_1_arg;
    begin_struct scardbegintransaction_1_arg;
    end_struct scardendtransaction_1_arg;
    cancel_struct scardcanceltransaction_1_arg;
    status_struct scardstatus_1_arg;
    statuschange_struct scardgetstatuschange_1_arg;
    transmit_struct scardtransmit_1_arg;
    cancellock_struct scardcancel_1_arg;
    list_struct scardlistreaders_1_arg;
    intro_struct scardintroducereader_1_arg;
    forget_struct scardforgetreader_1_arg;
  } argument;

  LONG rv;
  char *result;
  xdrproc_t xdr_argument, xdr_result;
  char *(*local)(char *, struct svc_req *);
  int pid;
  
  /* Zero out everything */
  result=0; rv=0; pid=0; local=0;

  /* Check the calling host to see if they are valid **********/
  rv = SECHVerifyHostValidity( transp );
  
  if ( rv != SCARD_S_SUCCESS ) {
    /* If the remote address is not allowed 
       to connect then send an error        */
    svcerr_systemerr(transp);
    return;
  }   
  /************************************************************/

  switch (rqstp->rq_proc) {
  case NULLPROC:
    (void) svc_sendreply(transp, (xdrproc_t) xdr_void, (char *)NULL);
    return;
    
  case SCARDESTABLISHCONTEXT:
    xdr_argument = (xdrproc_t) xdr_establish_struct;
    xdr_result = (xdrproc_t) xdr_establish_struct;
    local = (char *(*)(char *, struct svc_req *)) scardestablishcontext_1_svc;
    break;
    
  case SCARDRELEASECONTEXT:
    xdr_argument = (xdrproc_t) xdr_release_struct;
    xdr_result = (xdrproc_t) xdr_release_struct;
    local = (char *(*)(char *, struct svc_req *)) scardreleasecontext_1_svc;
    break;
    
  case SCARDCONNECT:
    xdr_argument = (xdrproc_t) xdr_connect_struct;
    xdr_result = (xdrproc_t) xdr_connect_struct;
    local = (char *(*)(char *, struct svc_req *)) scardconnect_1_svc;
    break;

  case SCARDRECONNECT:
    xdr_argument = (xdrproc_t) xdr_reconnect_struct;
    xdr_result = (xdrproc_t) xdr_reconnect_struct;
    local = (char *(*)(char *, struct svc_req *)) scardreconnect_1_svc;
    break;
    
  case SCARDDISCONNECT:
    xdr_argument = (xdrproc_t) xdr_disconnect_struct;
    xdr_result = (xdrproc_t) xdr_disconnect_struct;
    local = (char *(*)(char *, struct svc_req *)) scarddisconnect_1_svc;
    break;
    
  case SCARDBEGINTRANSACTION:
    xdr_argument = (xdrproc_t) xdr_begin_struct;
    xdr_result = (xdrproc_t) xdr_begin_struct;
    local = (char *(*)(char *, struct svc_req *)) scardbegintransaction_1_svc;
    break;
    
  case SCARDENDTRANSACTION:
    xdr_argument = (xdrproc_t) xdr_end_struct;
    xdr_result = (xdrproc_t) xdr_end_struct;
    local = (char *(*)(char *, struct svc_req *)) scardendtransaction_1_svc;
    break;
    
  case SCARDCANCELTRANSACTION:
    xdr_argument = (xdrproc_t) xdr_cancel_struct;
    xdr_result = (xdrproc_t) xdr_cancel_struct;
    local = (char *(*)(char *, struct svc_req *)) scardcanceltransaction_1_svc;
    break;
    
  case SCARDSTATUS:
    xdr_argument = (xdrproc_t) xdr_status_struct;
    xdr_result = (xdrproc_t) xdr_status_struct;
    local = (char *(*)(char *, struct svc_req *)) scardstatus_1_svc;
    break;
    
  case SCARDGETSTATUSCHANGE:
    xdr_argument = (xdrproc_t) xdr_statuschange_struct;
    xdr_result = (xdrproc_t) xdr_statuschange_struct;
    local = (char *(*)(char *, struct svc_req *)) scardgetstatuschange_1_svc;
    break;
    
  case SCARDTRANSMIT:
    xdr_argument = (xdrproc_t) xdr_transmit_struct;
    xdr_result = (xdrproc_t) xdr_transmit_struct;
    local = (char *(*)(char *, struct svc_req *)) scardtransmit_1_svc;
    break;
    
  case SCARDCANCEL:
    xdr_argument = (xdrproc_t) xdr_cancellock_struct;
    xdr_result = (xdrproc_t) xdr_cancellock_struct;
    local = (char *(*)(char *, struct svc_req *)) scardcancel_1_svc;
    break;
    
  case SCARDLISTREADERS:
    xdr_argument = (xdrproc_t) xdr_list_struct;
    xdr_result = (xdrproc_t) xdr_list_struct;
    local = (char *(*)(char *, struct svc_req *)) scardlistreaders_1_svc;
    break;

  case SCARDINTRODUCEREADER:
    xdr_argument = (xdrproc_t) xdr_intro_struct;
    xdr_result = (xdrproc_t) xdr_intro_struct;
    local = (char *(*)(char *, struct svc_req *)) scardintroducereader_1_svc;
    break;  

  case SCARDFORGETREADER:
    xdr_argument = (xdrproc_t) xdr_forget_struct;
    xdr_result = (xdrproc_t) xdr_forget_struct;
    local = (char *(*)(char *, struct svc_req *)) scardforgetreader_1_svc;
    break; 
    
  default:
    svcerr_noproc(transp);
    return;
  }

  (void) memset((char *)&argument, 0, sizeof (argument));

  if (!svc_getargs(transp, xdr_argument, (caddr_t) &argument)) {
    svcerr_decode(transp);
    return;
  }
  
  /* Addition of multi-process server */
  /* Parent goes back to listening    */
  /* Child does the work              */
  
  /* Make sure somebody is not trying to recursively spawn off 
     GetStatusChange threads in a malicious manner.  If the maximum
     is reached just ignore any new requests and return an error.
  */

  if ( rqstp->rq_proc == SCARDGETSTATUSCHANGE ) {    
    if ( dwProcessNum < PCSCLITE_MAX_THREADS ) {
      if ( (pid = SYS_Fork()) != 0 ) {
	dwProcessNum += 1;
	return;                
      }
    } else {
      svcerr_systemerr(transp);
      return;
    }    
  }
  
  result = (*local)((char *)&argument, rqstp);

  if (result != NULL && !svc_sendreply(transp, xdr_result, result)) {
    svcerr_systemerr(transp);
  }

  if (!svc_freeargs(transp, xdr_argument, (caddr_t) &argument)) {
    fprintf(stderr, "unable to free arguments");
    SYS_Exit(1);
  }
  
  if ( rqstp->rq_proc == SCARDGETSTATUSCHANGE ) {
    /* dwProcessNum -= 1; This won't work with fork() */
    SYS_Exit(0);
  }
  
  return;
}

int SVCReturnProcessNumber( void ) {
  return dwProcessNum;
}

