/******************************************************************

	MUSCLE SmartCard Development ( http://www.linuxnet.com )
	    Title  : winscard.c
	    Package: pcsc lite
            Author : David Corcoran
            Date   : 7/27/99
	    License: Copyright (C) 1999 David Corcoran
	             <corcoran@linuxnet.com>
            Purpose: This handles smartcard reader communications. 
	             This is the heart of the M$ smartcard API.
	            
********************************************************************/

#include <winscard.h>
#include <readerfactory.h>
#include <prothandler.h>
#include <ifdhandler.h>
#include <ifdwrapper.h>
#include <atrhandler.h>
#include <debuglog.h>
#include <configfile.h>
#include <sys_generic.h>
#include <stdlib.h>

/* Some defines for context stack */
#define SCARD_LAST_CONTEXT       1
#define SCARD_NO_CONTEXT         0
#define SCARD_EXCLUSIVE_CONTEXT -1
#define SCARD_NO_LOCK            0

SCARD_IO_REQUEST g_rgSCardT0Pci, g_rgSCardT1Pci, g_rgSCardRawPci;

LONG SCardEstablishContext( DWORD dwScope, LPCVOID pvReserved1, 
			    LPCVOID pvReserved2, LPSCARDCONTEXT phContext) {

  LONG rv;

  /* Zero out everything */
  rv=0;

  /* Unique identifier for this server so that it can uniquely 
     be identified by clients and distinguished from others    */

  *phContext = (PCSCLITE_SVC_IDENTITY + SYS_Random(0, 1.0, 65535.0));
  return SCARD_S_SUCCESS;
}

LONG SCardReleaseContext( SCARDCONTEXT hContext ) {
  /* Nothing to do here RPC layer will handle this */
  return SCARD_S_SUCCESS;
}

LONG SCardSetTimeout( SCARDCONTEXT hContext, DWORD dwTimeout ) {
  /* This is only used at the client side
     of an RPC call but just in case someone
     calls it here                           */

  return SCARD_E_UNSUPPORTED_FEATURE;
}

LONG SCardConnect( SCARDCONTEXT hContext, LPCSTR szReader, DWORD dwShareMode,
                   DWORD dwPreferredProtocols, LPSCARDHANDLE phCard,
                   LPDWORD pdwActiveProtocol ) {              
  LONG rv;
  PREADER_CONTEXT rContext; 
  UCHAR pucAtr[MAX_ATR_SIZE], ucAvailable;
  DWORD dwAtrLength, dwState, dwStatus;
  DWORD dwReaderLen, dwProtocol;

  /* Zero out everything */
  rv=0; rContext=0; ucAvailable=0;
  dwAtrLength=0; dwState=0; dwStatus=0; dwReaderLen=0; 
  dwProtocol=0; memset( pucAtr, 0x00, MAX_ATR_SIZE );

  /* Set hCard to zero */
  *phCard = 0;

#ifdef PCSC_DEBUG
    DebugLogC("SCardConnect: Attempting Connect to %s\n", szReader, 
	     __FILE__, __LINE__);
#endif

  rv = RFReaderInfo( (LPSTR)szReader, &rContext );   

  if ( rv != SCARD_S_SUCCESS ) { 
#ifdef PCSC_DEBUG
    DebugLogC("SCardConnect: Reader %s Not Found\n", szReader, 
	     __FILE__, __LINE__);
#endif
    return rv; 
  }

  /*******************************************/
  /* This section checks for simple errors   */
  /*******************************************/
  
  /* We only use T=0/T=1/RAW for now          */
  if ( dwShareMode != SCARD_SHARE_DIRECT ) {
    if ( (dwPreferredProtocols > 15)||(dwPreferredProtocols < 1 ) ) {
      return SCARD_E_UNSUPPORTED_FEATURE;   
    }
  }

  /* Connect if not exclusive mode    */
  if ( rContext->dwContexts == SCARD_EXCLUSIVE_CONTEXT ) {
#ifdef PCSC_DEBUG
    DebugLogA("SCardConnect: Error Reader Exclusive\n", __FILE__, __LINE__);
#endif
    return SCARD_E_SHARING_VIOLATION;
  }
  
  /*******************************************/
  /* This section tries to determine the     */
  /* presence of a card or not               */
  /*******************************************/      
  dwStatus = rContext->dwStatus;
  
  if ( dwShareMode != SCARD_SHARE_DIRECT ) {
    if ( !(dwStatus & SCARD_PRESENT) ) {      
#ifdef PCSC_DEBUG
      DebugLogA("SCardConnect: Card Not Inserted\n", __FILE__, __LINE__);
#endif
      return SCARD_E_NO_SMARTCARD;
    }
  }
      
  /*******************************************/
  /* This section tries to decode the ATR    */
  /* and set up which protocol to use        */
  /*******************************************/
  
  if ( dwShareMode != SCARD_SHARE_DIRECT ) {
    if ( rContext->dwContexts == SCARD_NO_CONTEXT ) {    
      memcpy( pucAtr, rContext->ucAtr, rContext->dwAtrLen );
      dwAtrLength = rContext->dwAtrLen;
    
      rContext->dwProtocol = PHGetDefaultProtocol( pucAtr, dwAtrLength );
      ucAvailable          = PHGetAvailableProtocols( pucAtr, dwAtrLength );
      rContext->dwProtocol = PHSetProtocol( rContext, dwPreferredProtocols,
	  				    ucAvailable );
    }
  }
  
  *pdwActiveProtocol = rContext->dwProtocol; 
  
#ifdef PCSC_DEBUG
  DebugLogB("SCardConnect: Active Protocol: %d\n", *pdwActiveProtocol, 
	    __FILE__, __LINE__);
#endif
    
  /* Prepare the SCARDHANDLE identity  */
  *phCard = RFCreateReaderHandle( rContext );
  
#ifdef PCSC_DEBUG
  DebugLogB("SCardConnect: hCard Identity: %x\n", *phCard, 
	    __FILE__, __LINE__);
#endif
  
  /*******************************************/
  /* This section tries to set up the        */
  /* exclusivity modes. -1 is exclusive      */
  /*******************************************/

  if ( dwShareMode == SCARD_SHARE_EXCLUSIVE ) {
    if ( rContext->dwContexts == SCARD_NO_CONTEXT ) {
      rContext->dwContexts = SCARD_EXCLUSIVE_CONTEXT;
      RFLockSharing( *phCard );
    } else {
      RFDestroyReaderHandle( *phCard );
      *phCard = 0;
      return SCARD_E_SHARING_VIOLATION;
    }
  } else {
    /* Add a connection to the context stack */
    rContext->dwContexts += 1;
  }
  
  /* Add this handle to the handle list */
  RFAddReaderHandle( rContext, *phCard );

  return SCARD_S_SUCCESS;
}

LONG SCardReconnect( SCARDHANDLE hCard, DWORD dwShareMode,
		     DWORD dwPreferredProtocols, DWORD dwInitialization,
		     LPDWORD pdwActiveProtocol ) {              
  LONG rv;
  PREADER_CONTEXT rContext; 
  UCHAR pucAtr[MAX_ATR_SIZE], ucAvailable;
  DWORD dwAtrLength, dwAction;
  DWORD dwReaderLen, dwProtocol;

  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }
  
  rv = RFReaderInfoById( hCard, &rContext );   
  if ( rv != SCARD_S_SUCCESS ) { return rv; }
  
  /* Handle the dwShareMode */

  if ( dwShareMode == SCARD_SHARE_EXCLUSIVE ) {
    if ( rContext->dwContexts == SCARD_EXCLUSIVE_CONTEXT ) {
      /* Do nothing - we are already exclusive */
    } else {
      if ( rContext->dwContexts == SCARD_LAST_CONTEXT ) {
	/* Make sure we are only connected then exclusify */
	rContext->dwContexts = SCARD_EXCLUSIVE_CONTEXT;
	RFLockSharing( hCard );
      } else {
	return SCARD_E_SHARING_VIOLATION;
      }      
    }
  } else if ( dwShareMode == SCARD_SHARE_SHARED ) {
    if ( rContext->dwContexts != SCARD_EXCLUSIVE_CONTEXT ) {
      /* Do nothing - in sharing mode already */
    } else {
      /* We are in exclusive mode but want to share now */
      RFUnlockSharing( hCard );
      rContext->dwContexts = SCARD_LAST_CONTEXT;
    }
  } else {
    return SCARD_E_INVALID_VALUE;
  }

  /* Handle the dwInitialization */

  switch( dwInitialization ) {
  case SCARD_LEAVE_CARD:     /* Do nothing here */
    break;                    
  case SCARD_UNPOWER_CARD:
    dwAction = IFD_POWER_DOWN;
    break;
  case SCARD_RESET_CARD:
    dwAction = IFD_RESET;
    break;
  case SCARD_EJECT_CARD:
    break;
  default:
    return SCARD_E_INVALID_VALUE;
  };
  
  /* Thread handles powering so just reset instead */
  if ( dwInitialization == SCARD_UNPOWER_CARD ) {
    dwAction = IFD_RESET;
  }
  
  /* RFUnblockReader( rContext ); FIX - this doesn't work */
  
  /* Allow RESET only if no other application holds a lock */
  
  /*  if ( RFCheckSharing( hCard ) == SCARD_S_SUCCESS ) {    */
  if ( dwInitialization != SCARD_LEAVE_CARD && 
       dwInitialization != SCARD_EJECT_CARD ) 
    {	  

      /* Not doing this could result in deadlock */
      if ( RFCheckReaderEventState(rContext, hCard) != SCARD_W_RESET_CARD ) {
	rv = IFDPowerICC( rContext, dwAction, rContext->ucAtr, 
			  &rContext->dwAtrLen );

      /* Notify the card has been reset */
      RFSetReaderEventState( rContext, SCARD_RESET );
      
#ifdef PCSC_DEBUG
      if ( rContext->dwAtrLen > 0 ) {
	DebugLogA("SCardReconnect: Reset complete.\n", 
		  __FILE__, __LINE__ );
      } else {
	DebugLogA("SCardReconnect: Error resetting card.\n", 
		  __FILE__, __LINE__);
      }	 
#endif	 
      }
      
      /*      }   */
    }   
  
  /* Handle the dwActive/Preferred Protocols */
  
  memcpy( pucAtr, rContext->ucAtr, rContext->dwAtrLen );
  dwAtrLength = rContext->dwAtrLen;
  
  rContext->dwProtocol = PHGetDefaultProtocol( pucAtr, dwAtrLength );
  ucAvailable          = PHGetAvailableProtocols( pucAtr, dwAtrLength );
  rContext->dwProtocol = PHSetProtocol( rContext, dwPreferredProtocols,
					ucAvailable );
  
  *pdwActiveProtocol = rContext->dwProtocol; 
  
  if ( dwShareMode == SCARD_SHARE_EXCLUSIVE ) {
    if ( rContext->dwContexts == SCARD_EXCLUSIVE_CONTEXT ) {
      /* Do nothing - we are already exclusive */
    } else {
      if ( rContext->dwContexts == SCARD_LAST_CONTEXT ) {
	rContext->dwContexts = SCARD_EXCLUSIVE_CONTEXT;
	RFLockSharing( hCard );
      } else {
	return SCARD_E_SHARING_VIOLATION;
      }      
    }
  } else if ( dwShareMode == SCARD_SHARE_SHARED ) {
    if ( rContext->dwContexts != SCARD_EXCLUSIVE_CONTEXT ) {
      /* Do nothing - in sharing mode already */
    } else {
      /* We are in exclusive mode but want to share now */
      RFUnlockSharing( hCard );
      rContext->dwContexts = SCARD_LAST_CONTEXT;
    }
  } else {
    return SCARD_E_INVALID_VALUE;
  }
    
  /* Clear a previous event to the application */
  RFClearReaderEventState( rContext, hCard );

  return SCARD_S_SUCCESS;
}

LONG SCardDisconnect( SCARDHANDLE hCard, DWORD dwDisposition ) {
  
  LONG rv;
  LPBYTE pucAtr[MAX_ATR_SIZE];
  PREADER_CONTEXT rContext; 
  DWORD dwAction, dwAtrLen;

  /* Zero out everything */
  rv=0; rContext=0; dwAction=0; dwAtrLen=0;
  memset( pucAtr, 0x00, MAX_ATR_SIZE );

  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }
  
  rv = RFReaderInfoById( hCard, &rContext );   
  if ( rv != SCARD_S_SUCCESS ) { return rv; }
  
  switch( dwDisposition ) {
  case SCARD_LEAVE_CARD:     /* Do nothing here */
    break;                    
  case SCARD_UNPOWER_CARD:
    dwAction = IFD_POWER_DOWN;
    break;
  case SCARD_RESET_CARD:
    dwAction = IFD_RESET;
    break;
  case SCARD_EJECT_CARD:
    break;
  default:
    return SCARD_E_INVALID_VALUE;
  };
  
  /* Thread handles powering so just reset instead */
  if ( dwDisposition == SCARD_UNPOWER_CARD ) {
    dwAction = IFD_RESET;
  }

  /* Unlock any blocks on this context */
  RFUnlockSharing( hCard );
  
#ifdef PCSC_DEBUG
  DebugLogB("SCardDisconnect: Active Contexts: %d\n", rContext->dwContexts, 
	    __FILE__, __LINE__);
#endif
  
  /* RFUnblockReader( rContext ); FIX - this doesn't work */
  
  /* Allow RESET only if no other application holds a lock */

  /*  Deprecated - any app can reset according to M$       */
  /*  if ( RFCheckSharing( hCard ) == SCARD_S_SUCCESS ) {  */
    if ( dwDisposition != SCARD_LEAVE_CARD && 
	 dwDisposition != SCARD_EJECT_CARD ) 
      {	  
	rv = IFDPowerICC( rContext, dwAction, rContext->ucAtr, 
			  &rContext->dwAtrLen );
	
	/* Notify the card has been reset */
	RFSetReaderEventState( rContext, SCARD_RESET );
	
#ifdef PCSC_DEBUG
	if ( rContext->dwAtrLen > 0 ) {
	  DebugLogA("SCardDisconnect: Reset complete.\n", 
		    __FILE__, __LINE__ );
	} else {
	  DebugLogA("SCardDisconnect: Error resetting card.\n", 
		    __FILE__, __LINE__);
	}	 
#endif	 
	  /*  Deprecated - any app can reset according to M$       */
	  /*   }                                                   */
  }   
  
  /* Remove and destroy this handle */
  RFRemoveReaderHandle( rContext, hCard );
  RFDestroyReaderHandle( hCard );

  /* For exclusive connection reset it to no connections */
  if ( rContext->dwContexts == SCARD_EXCLUSIVE_CONTEXT ) {
    rContext->dwContexts = SCARD_NO_CONTEXT;
    return SCARD_S_SUCCESS;
  }

  /* Remove a connection from the context stack */
  rContext->dwContexts -= 1;
  return SCARD_S_SUCCESS;
}

LONG SCardBeginTransaction( SCARDHANDLE hCard ) {

  LONG rv;
  PREADER_CONTEXT rContext;

  /* Zero out everything */
  rv=0;

  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }
  
  rv = RFReaderInfoById( hCard, &rContext );   

  /* Cannot find the hCard in this context */
  if ( rv != SCARD_S_SUCCESS ) { 
    return rv; 
  }

  /* Make sure some event has not occurred */
  if ( (rv = RFCheckReaderEventState( rContext, hCard )) 
       != SCARD_S_SUCCESS ) 
    {
      return rv;
    }

  rv = RFLockSharing( hCard );

#ifdef PCSC_DEBUG
  DebugLogB("SCardBeginTransaction: Status: %d.\n", rv,
            __FILE__, __LINE__);
#endif

  return rv; 
}

LONG SCardEndTransaction( SCARDHANDLE hCard, DWORD dwDisposition ) {

  LONG rv;
  PREADER_CONTEXT rContext;

  /* Zero out everything */
  rv=0;

  /* Ignoring dwDisposition for now */
  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }

  rv = RFReaderInfoById( hCard, &rContext );   

  /* Cannot find the hCard in this context */
  if ( rv != SCARD_S_SUCCESS ) { 
    return rv; 
  }

  /* Make sure some event has not occurred */
  if ( (rv = RFCheckReaderEventState( rContext, hCard )) 
       != SCARD_S_SUCCESS ) 
    {
      return rv;
    }
  
  rv = RFUnlockSharing( hCard );  

#ifdef PCSC_DEBUG
  DebugLogB("SCardEndTransaction: Status: %d.\n", rv,
            __FILE__, __LINE__);
#endif

  return rv;
}

LONG SCardCancelTransaction( SCARDHANDLE hCard ) {

  LONG rv;
  PREADER_CONTEXT rContext;

  /* Zero out everything */
  rv=0;

  /* Ignoring dwDisposition for now */
  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }
  
  rv = RFReaderInfoById( hCard, &rContext );   
  
  /* Cannot find the hCard in this context */
  if ( rv != SCARD_S_SUCCESS ) {
    return rv; 
  }
  
  /* Make sure some event has not occurred */
  if ( (rv = RFCheckReaderEventState( rContext, hCard )) 
       != SCARD_S_SUCCESS ) 
    {
      return rv;
    }
  
  rv = RFUnlockSharing( hCard );  

#ifdef PCSC_DEBUG
  DebugLogB("SCardCancelTransaction: Status: %d.\n", rv,
            __FILE__, __LINE__);
#endif

  return rv;
}

LONG SCardStatus( SCARDHANDLE hCard, LPSTR mszReaderNames, 
		  LPDWORD pcchReaderLen, LPDWORD pdwState, 
		  LPDWORD pdwProtocol, LPBYTE pbAtr,
                  LPDWORD pcbAtrLen ) {  
  LONG rv;
  LONG lReaderSize;
  PREADER_CONTEXT rContext; 

  /* Zero out everything */
  rv=0; lReaderSize=0; rContext=0;

  *pcbAtrLen      = 0;
  *pdwState       = 0;
  *pdwProtocol    = 0;

  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }
  
  rv = RFReaderInfoById( hCard, &rContext );   

  /* Cannot find the hCard in this context */
  if ( rv != SCARD_S_SUCCESS ) { 
    *pcchReaderLen  = 0;
    return rv; 
  }

  /* Make sure some event has not occurred */
  if ( (rv = RFCheckReaderEventState( rContext, hCard )) 
       != SCARD_S_SUCCESS ) 
    {
      return rv;
    }
  
  /* Check to see if the buffer is large enough
     if not return the correct size of the buffer */
  
  lReaderSize = strlen( rContext->lpcReader ) + 1;
  if ( *pcchReaderLen < lReaderSize || mszReaderNames == 0 ) 
    {
      *pcchReaderLen = lReaderSize;
      return SCARD_E_INSUFFICIENT_BUFFER;
    }
  
  *pdwState    = rContext->dwStatus;
  *pdwProtocol = rContext->dwProtocol;
  if ( *pdwState & SCARD_PRESENT ) {
    *pcbAtrLen = rContext->dwAtrLen;
    memcpy( pbAtr, rContext->ucAtr, *pcbAtrLen );
  } else {
    *pcbAtrLen = 0;
  }
  
  /* Return the structure values */
  strcpy( mszReaderNames, rContext->lpcReader );
  *pcchReaderLen  = lReaderSize;
  *pdwProtocol    = rContext->dwProtocol;
  
  return SCARD_S_SUCCESS;
}

LONG SCardGetStatusChange( SCARDCONTEXT hContext, DWORD dwTimeout, 
			   LPSCARD_READERSTATE_A rgReaderStates, 
                           DWORD cReaders ) {

  LONG rv;
  PREADER_CONTEXT rContext; 
  LPSTR lpcReaderName;
  DWORD dwEventState, dwAtrSize;
  BYTE pbAtr[MAX_ATR_SIZE];
  DWORD dwCurrentState, dwTime, dwProtocol;
  DWORD dwActiveProtocol, dwState;
  SCARDHANDLE hCard;

  /* Zero out everything */
  rv=0; rContext=0; lpcReaderName=0; dwEventState=0; dwAtrSize=0;
  dwCurrentState=0; dwTime=0; dwProtocol=0; dwActiveProtocol=0;
  dwState=0; hCard=0; memset( pbAtr, 0x00, MAX_ATR_SIZE );

  dwCurrentState               = rgReaderStates->dwCurrentState;
  rgReaderStates->dwEventState = 0;
  dwEventState                 = 0;
  dwTime                       = 0;

  /* Just ignore this one and return */
  if ( rgReaderStates->dwCurrentState & SCARD_STATE_IGNORE ) {
    rgReaderStates->dwEventState = SCARD_STATE_IGNORE;
    return SCARD_S_SUCCESS;
  }

  lpcReaderName = (char *)rgReaderStates->szReader;
  
/*  if ( strcmp( lpcReaderName, "$SCardAllReaders" ) == 0 ) { */
  
  rv = RFReaderInfo( lpcReaderName, &rContext );   
  if ( rv != SCARD_S_SUCCESS ) {
    rgReaderStates->dwEventState = 
      SCARD_STATE_UNKNOWN | SCARD_STATE_CHANGED | SCARD_STATE_IGNORE;   
    return rv; 
  }
  
  /* Now is where we start our event checking loop */
  
#ifdef PCSC_DEBUG
  DebugLogA("SCardGetStatusChange: Event Loop Start\n", __FILE__, __LINE__);
#endif
  
  rv = RFReaderInfo( (char *)lpcReaderName, &rContext );   
  if ( rv != SCARD_S_SUCCESS ) { 
    rgReaderStates->dwEventState = 
      SCARD_STATE_UNKNOWN | SCARD_STATE_CHANGED | SCARD_STATE_IGNORE;
    return rv; 
  } else if ( dwCurrentState & SCARD_STATE_UNAVAILABLE ) {
    dwEventState |= SCARD_STATE_CHANGED;
  }
  
  do {
    
    /* Now we check all the Reader States */
    dwState   = rContext->dwStatus;
    if ( dwState & SCARD_PRESENT ) {
      dwAtrSize = rContext->dwAtrLen;
      memcpy( pbAtr, rContext->ucAtr, dwAtrSize );
    } else {
      dwAtrSize = 0;
    }

    /* Card is now absent                   */
    if ( dwState & SCARD_ABSENT ) 
      {
	dwEventState |= SCARD_STATE_EMPTY;

	/* After present the rest are assumed */
	if ( dwCurrentState & SCARD_STATE_PRESENT ||
	     dwCurrentState & SCARD_STATE_ATRMATCH ||
	     dwCurrentState & SCARD_STATE_EXCLUSIVE ||
	     dwCurrentState & SCARD_STATE_INUSE ) 
	  {
	    dwEventState |= SCARD_STATE_CHANGED;	
	  }
	
	/* Card is now present              */
      } else {
	dwEventState |= SCARD_STATE_PRESENT;
	
	if ( dwCurrentState & SCARD_STATE_EMPTY ) {
	  dwEventState |= SCARD_STATE_CHANGED;
	}	
      }
	/* Now figure out sharing modes     */
    if ( rContext->dwContexts == -1 ) {
      dwEventState |= SCARD_STATE_EXCLUSIVE;
      if ( !dwCurrentState & SCARD_STATE_EXCLUSIVE ) {
	dwEventState |= SCARD_STATE_CHANGED;
      }
    } else if ( rContext->dwContexts >= 1 ) {
      dwEventState |= SCARD_STATE_INUSE;
      if ( !dwCurrentState & SCARD_STATE_INUSE ) {
	dwEventState |= SCARD_STATE_CHANGED;
      }
    }
    
    SYS_USleep( PCSCLITE_STATUS_WAIT );
    
    if ( dwTimeout != INFINITE ) {
      dwTime += PCSCLITE_STATUS_WAIT;
      
      if ( dwTime >= (dwTimeout*1000) ) {
	rgReaderStates->cbAtr        = dwAtrSize;
	rgReaderStates->dwEventState = dwEventState;
        memcpy( rgReaderStates->rgbAtr, pbAtr, MAX_ATR_SIZE );
        return SCARD_E_TIMEOUT;
      }
    }
    
  } while ( (dwEventState & SCARD_STATE_CHANGED) == 0   &&
	    (dwCurrentState != SCARD_STATE_UNAWARE)     &&
	    (dwTimeout != 0)                            &&
	    rContext->dwBlockStatus != hContext );

#ifdef PCSC_DEBUG
    DebugLogA("SCardGetStatusChange: Event Loop End\n", __FILE__, __LINE__);
#endif

    rgReaderStates->cbAtr        = dwAtrSize;
    rgReaderStates->dwEventState = dwEventState;
    memcpy( rgReaderStates->rgbAtr, pbAtr, MAX_ATR_SIZE );

    return SCARD_S_SUCCESS;
}

LONG SCardControl( SCARDHANDLE hCard, LPCBYTE pbSendBuffer, DWORD cbSendLength,
                   LPBYTE pbRecvBuffer, LPDWORD pcbRecvLength ) {

    /* This is not used.  SCardControl is passed through SCardTransmit.
       This is here to make the compiler happy.
    */
       
    return SCARD_S_SUCCESS;
}

LONG SCardTransmit( SCARDHANDLE hCard, LPCSCARD_IO_REQUEST pioSendPci,
                    LPCBYTE pbSendBuffer, DWORD cbSendLength,
                    LPSCARD_IO_REQUEST pioRecvPci, LPBYTE pbRecvBuffer, 
		    LPDWORD pcbRecvLength ) {
 
  LONG rv;
  PREADER_CONTEXT rContext; 
  SCARD_IO_HEADER sSendPci, sRecvPci;
  DWORD dwRxLength;

  /* Zero out everything */
  rv=0; rContext=0; dwRxLength=0;

  dwRxLength     = *pcbRecvLength;
  *pcbRecvLength = 0;

  if ( hCard == 0 ) { return SCARD_E_INVALID_HANDLE; }

  /* Make sure no one has a lock on this reader */
  if ( (rv = RFCheckSharing(hCard)) != SCARD_S_SUCCESS ) {
    return rv;
  }

  rv = RFReaderInfoById( hCard, &rContext );   
  if ( rv != SCARD_S_SUCCESS ) { return rv; }

  /* Make sure some event has not occurred */
  if ( (rv = RFCheckReaderEventState( rContext, hCard )) 
       != SCARD_S_SUCCESS ) 
    {
      return rv;
    }

  /* Check for some common errors */

  if ( rContext->dwStatus & SCARD_ABSENT ) {
    return SCARD_E_NO_SMARTCARD;
  }

  if ( pioSendPci->dwProtocol != SCARD_PROTOCOL_RAW ) {
    if ( pioSendPci->dwProtocol != rContext->dwProtocol ) {
      return SCARD_E_PROTO_MISMATCH;
    }
  }

  if ( cbSendLength > MAX_BUFFER_SIZE ) {
    return SCARD_E_INSUFFICIENT_BUFFER;
  }

  if ( dwRxLength > MAX_BUFFER_SIZE ) {
    return SCARD_E_INSUFFICIENT_BUFFER;
  }

  /* Quick fix: PC/SC starts at 1 for bit masking
     but the IFD_Handler just wants 0 or 1        */

  if ( pioSendPci->dwProtocol == SCARD_PROTOCOL_T0 ) {
    sSendPci.Protocol = 0;
  } else if ( pioSendPci->dwProtocol == SCARD_PROTOCOL_T1 ) {
    sSendPci.Protocol = 1;
  } else if ( pioSendPci->dwProtocol == SCARD_PROTOCOL_RAW ) {
    /* This is temporary ...... */
    sSendPci.Protocol = SCARD_PROTOCOL_RAW;
  }

  sSendPci.Length     = pioSendPci->cbPciLength;

#ifdef PCSC_DEBUG
  DebugLogB("SCardTransmit: Send Protocol: %d\n", sSendPci.Protocol, 
	    __FILE__, __LINE__);
#endif
  
  if ( pioSendPci->dwProtocol == SCARD_PROTOCOL_RAW ) {
    rv = IFDControl( rContext, (PUCHAR)pbSendBuffer, cbSendLength, 
                     pbRecvBuffer, &dwRxLength );
  } else {
    rv = IFDTransmit( rContext, sSendPci, (PUCHAR)pbSendBuffer, cbSendLength,
                      pbRecvBuffer, &dwRxLength, &sRecvPci );
  }
  pioRecvPci->dwProtocol  = sRecvPci.Protocol;
  pioRecvPci->cbPciLength = sRecvPci.Length;

  /* Check for any errors that might have occurred */

  if ( rv != SCARD_S_SUCCESS ) { 
    *pcbRecvLength = 0;
    return SCARD_E_NOT_TRANSACTED;
  }

  if ( (dwRxLength <= MAX_BUFFER_SIZE) && (dwRxLength >= 0) ) {
    *pcbRecvLength          = dwRxLength;
  } else {
    *pcbRecvLength          = 0;
    return SCARD_E_INSUFFICIENT_BUFFER;
  }  
  
  return SCARD_S_SUCCESS;
}

LONG SCardListReaders( SCARDCONTEXT hContext, LPCSTR mszGroups, 
		       LPSTR mszReaders, LPDWORD pcchReaders ) {
  LONG rv;

  /* Zero out everything */
  rv=0;

  rv = RFListReaders( mszReaders, pcchReaders );
  return rv;
}

LONG SCardCancel( SCARDCONTEXT hContext ) {

  LONG rv;

  /* Zero out everything */
  rv=0;

  /* Issue a cancel request to the
     GetStatusChange function      */

  rv = RFUnblockContext( hContext );
  return rv;
}

